"""
This module provides structures for performing data labelling on a background image
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import time
from collections import OrderedDict
import json
from typing import Union, List, Tuple, Any, Dict, Optional
from datetime import datetime
import getpass

from sarpy.annotation.base import AnnotationProperties, FileAnnotationCollection, \
    AnnotationFeature, AnnotationCollection
from sarpy.geometry.geometry_elements import Jsonable, Geometry, Point, MultiPoint, \
    LineString, MultiLineString, Polygon, MultiPolygon, GeometryCollection

_LABEL_VERSION = "Label:1.0"
logger = logging.getLogger(__name__)
POSSIBLE_GEOMETRIES = ('point', 'line', 'polygon')


class LabelSchema(object):
    """
    The basic structure for an annotation/labelling schema.

    The label names may certainly be modified in place through use of the `labels`
    property, without worry for causing errors. This is discouraged, because having two
    schemas with same version number/ids and differing names can likely lead to confusion.

    Any modification of label ids of sub-id structure should be performed by using the
    :func:`set_labels_and_subtypes` method, or difficult to diagnose runtime errors
    will likely be introduced.
    """

    __slots__ = (
        '_version', '_labels', '_classification', '_version_date', '_subtypes',
        '_parent_types', '_confidence_values', '_permitted_geometries',
        '_integer_ids', '_maximum_id')

    def __init__(
            self,
            version: Optional[str] = '1.0',
            labels: Optional[Dict[str, str]] = None,
            version_date: Optional[str] = None,
            classification: str = "UNCLASSIFIED",
            subtypes: Optional[Dict[str, List[str]]] = None,
            confidence_values: Optional[List[Union[int, str]]] = None,
            permitted_geometries: Optional[List[str]] = None):
        """

        Parameters
        ----------
        version : None|str
            The version of the schema.
        labels : None|Dict[str, str]
            The {<label id> : <label name>} pair dictionary. Each entry must be a string,
            and '' is not a valid label id.
        version_date : None|str
            The date for this schema. If `None`, then the current time will be used.
        classification : str
            The classification for this schema.
        subtypes : None|Dict[str, List[str]]
            The {<label id> : <sub id list>} pairs. The root ids (i.e. those ids
            not belonging as sub-id to some other id) will be populated in the subtypes
            entry with empty string key (i.e. ''). Every key and entry of subtypes
            (excluding the subtypes root '') must correspond to an entry of labels,
            and no id can be a direct subtype of more than one id.
        confidence_values : None|List[Union[str, int]]
            The possible confidence values.
        permitted_geometries : None|List[str]
            The possible geometry types.
        """

        self._version_date = None
        self._labels = None
        self._subtypes = None
        self._parent_types = None
        self._confidence_values = None
        self._permitted_geometries = None
        self._integer_ids = True
        self._maximum_id = None  # type: Union[None, int]

        self._version = version
        self.update_version_date(value=version_date)
        self._classification = classification

        self.confidence_values = confidence_values
        self.permitted_geometries = permitted_geometries
        self.set_labels_and_subtypes(labels, subtypes)

    @property
    def version(self) -> str:
        """
        The version of the schema.

        Returns
        -------
        str
        """

        return self._version

    @property
    def version_date(self) -> str:
        """
        The date for this schema version - this should be a viable datetime format,
        but this is unenforced.

        Returns
        -------
        str
        """

        return self._version_date

    def update_version_date(self, value: Optional[str] = None):
        if isinstance(value, str):
            self._version_date = value
        else:
            self._version_date = datetime.utcnow().isoformat('T')+'Z'

    @property
    def classification(self) -> str:
        """
        str: The classification for the contents of this schema.
        """

        return self._classification

    @property
    def suggested_next_id(self) -> Optional[int]:
        """
        None|int: If all ids are integer type, this returns max_id+1. Otherwise, this
        yields None.
        """

        return None if self._maximum_id is None else self._maximum_id + 1

    @property
    def labels(self) -> Dict[str, str]:
        """
        The complete label dictionary of the form `{label_id : label_name}`.

        Returns
        -------
        Dict[str, str]
        """

        return self._labels

    @property
    def subtypes(self) -> Dict[str, List[str]]:
        """
        The complete dictionary of subtypes of the form `{parent_id : <subids list>}`.

        Returns
        -------
        Dict[str, List[str]]
        """

        return self._subtypes

    @property
    def parent_types(self) -> Dict[str, List[str]]:
        """
        The dictionary of parent types of the form `{child_id : <set of parent ids>}`.
        It is canonically defined that an id is a parent of itself. The order of
        the `parent_ids` list is ascending order of parentage, i.e.
        `[<self>, <parent>, <parent of parent>, ...]`.

        Returns
        -------
        Dict[str, List[str]]
        """

        return self._parent_types

    @property
    def confidence_values(self) -> List[Union[int, str]]:
        """
        The list of confidence values.

        Returns
        -------
        List
            Each element should be a json type (most likely use cases are str or int).
        """

        return self._confidence_values

    @confidence_values.setter
    def confidence_values(self, conf_values):
        if conf_values is None:
            self._confidence_values = None
            return

        if not isinstance(conf_values, list):
            raise TypeError('confidence_values must be a list. Got type {}'.format(type(conf_values)))
        self._confidence_values = conf_values

    @property
    def permitted_geometries(self) -> Optional[List[str]]:
        """
        The collection of permitted geometry types. None corresponds to all.
        Entries should be one of `{'point', 'line', 'polygon'}`.

        Returns
        -------
        None|List[str]
        """

        return self._permitted_geometries

    @permitted_geometries.setter
    def permitted_geometries(self, values):
        if values is None:
            self._permitted_geometries = None
            return

        if isinstance(values, str):
            values = [values.lower().strip(), ]
        else:
            values = [entry.lower().strip() for entry in values]

        if len(values) == 0:
            self._permitted_geometries = None
            return

        temp_values = []
        for entry in values:
            if entry in temp_values:
                continue
            if entry not in POSSIBLE_GEOMETRIES:
                raise ValueError('Got unknown geometry value `{}`'.format(entry))
            temp_values.append(entry)

        self._permitted_geometries = temp_values

    def get_id_from_name(self, the_name: str) -> Optional[str]:
        """
        Determine the id from the given name. Get `None` if this fails.

        Parameters
        ----------
        the_name : str

        Returns
        -------
        None|str
        """

        prospective = None
        for key, value in self.labels.items():
            if value == the_name:
                prospective = key
                break
        return prospective

    def get_parent(self, the_id: str) -> str:
        """
        Get the parent id for the given element id. The empty string is returned
        for elements with no parent.

        Parameters
        ----------
        the_id : str

        Returns
        -------
        str
        """

        parents = self.parent_types[the_id]
        return parents[1] if len(parents) > 1 else ''

    def __str__(self) -> str:
        return json.dumps(self.to_dict(), indent=1)

    def __repr__(self) -> str:
        return json.dumps(self.to_dict())

    def _inspect_new_id_for_integer(self, the_id: Union[int, str]) -> None:
        if not self._integer_ids:
            return  # nothing to do
        if isinstance(the_id, str):
            # noinspection PyBroadException
            try:
                the_id = int(the_id)
            except Exception:
                self._integer_ids = False
                self._maximum_id = None

        if isinstance(the_id, int):
            # noinspection PyTypeChecker
            self._maximum_id = the_id if self._maximum_id is None else \
                max(self._maximum_id, the_id)
        else:
            self._integer_ids = False
            self._maximum_id = None

    def _inspect_ids_for_integer(self) -> None:
        for the_id in self._labels:
            self._inspect_new_id_for_integer(the_id)

    @staticmethod
    def _find_inverted_fork(
            subtypes: Dict[str, List[str]],
            labels: Dict[str, str]) -> Dict[str, List[str]]:
        """
        Look for parents claiming the same child. This assigns all unclaimed children
        to '' parent.

        Parameters
        ----------
        subtypes : dict
        labels : dict

        Returns
        -------
        dict
        """

        # we need to check the reference count for each key in labels
        counts = OrderedDict((key, 0) for key in labels)

        # ensure that every key of subtypes is a string and every value is a list,
        # also that inclusion makes sense
        for key, value in subtypes.items():
            if not isinstance(key, str):
                raise TypeError(
                    'All keys of subtypes must be of type string. Got key `{}` of '
                    'type {}.'.format(key, type(key)))
            if key != '' and key not in labels:
                raise KeyError(
                    'All keys of subtypes must belong to labels. Got key `{}` '
                    'which is missing from labels.'.format(key))

            if not isinstance(value, list):
                raise TypeError(
                    'All values of subtypes must be of type `list`. Got value {} '
                    'for key `{}` of type {}'.format(value, key, type(value)))
            for entry in value:
                if entry not in labels:
                    raise KeyError(
                        'All entries for each value of subtypes must belong to labels. '
                        'Got entry `{}` in key `{}` which is missing from labels.'.format(entry, key))
                counts[entry] += 1
        # create the root entry for subtypes
        if '' not in subtypes:
            subtypes[''] = []
        if isinstance(subtypes, OrderedDict):
            subtypes.move_to_end('', last=False)
        for key in counts:
            value = counts[key]
            if value > 1:
                raise ValueError('key {} is referenced in more than one subtype. This is invalid.'.format(key))
            if value == 0:
                subtypes[''].append(key)
        return subtypes

    @staticmethod
    def _find_cycle(subtypes: Dict[str, List[str]]) -> None:
        """
        Find any cycles in the data.

        Parameters
        ----------
        subtypes

        Returns
        -------
        None
        """

        found_cycles = []

        def iterate(current_id, find_id):
            for t_entry in subtypes.get(current_id, []):
                if t_entry == find_id:
                    found_cycles.append((find_id, current_id))
                iterate(t_entry, find_id)

        for the_id in subtypes['']:
            iterate(the_id, the_id)
        if len(found_cycles) > 0:
            for entry in found_cycles:
                logger.error(
                    'Cycle found with ids {} and {}'.format(entry[0], entry[1]))
            raise ValueError('cycles found in graph information')

    def set_labels_and_subtypes(
            self,
            labels: Dict[str, str],
            subtypes: Dict[str, List[str]]) -> None:
        """
        Set the labels and subtypes. **Note that subtypes may be modified in place.**

        Parameters
        ----------
        labels : None|dict
        subtypes : None|dict

        Returns
        -------
        None
        """

        if labels is None:
            labels = OrderedDict()
        if not isinstance(labels, dict):
            raise TypeError('labels is required to be a dict. Got type {}'.format(type(labels)))

        if subtypes is None:
            subtypes = OrderedDict()
        elif not isinstance(subtypes, dict):
            raise TypeError('subtypes is required to be None or a dict. Got type {}'.format(type(subtypes)))

        # ensure that every key and value of labels are strings
        for key in labels:
            if not isinstance(key, str):
                raise TypeError(
                    'All keys of labels must be of type string. Got key `{}` of '
                    'type {}'.format(key, type(key)))
            if key == '':
                raise ValueError('The empty string is not a valid label id.')
            value = labels[key]
            if not isinstance(value, str):
                raise TypeError(
                    'All values of labels must be of type string. Got value {} '
                    'for key `{}` of type {}'.format(value, key, type(value)))

        # look for inverted fork - multiple parents claiming the same child
        subtypes = self._find_inverted_fork(subtypes, labels)
        # look for cycles
        self._find_cycle(subtypes)

        # set the values
        self._labels = labels
        self._subtypes = subtypes
        self._construct_parent_types()
        self._inspect_ids_for_integer()

    def _construct_parent_types(self) -> None:
        def iterate(t_key, parents):
            entry = [t_key, ]
            # noinspection PyUnresolvedReferences
            entry.extend(parents)
            self._parent_types[t_key] = entry
            if t_key not in self._subtypes:
                return

            for child_key in self._subtypes[t_key]:
                iterate(child_key, entry)

        self._parent_types = {}
        for key in self._subtypes['']:
            iterate(key, [])

    def _validate_entry(
            self,
            the_id: str,
            the_name: str,
            the_parent: str) -> Tuple[str, str, str]:
        """
        Validate the basics for the given entry.

        Parameters
        ----------
        the_id : str
        the_name : str
        the_parent : str

        Returns
        -------
        the_id: str
        the_name: str
        the_parent: str
        """

        # validate inputs
        if not (isinstance(the_id, str) and isinstance(the_name, str) and isinstance(the_parent, str)):
            raise TypeError(
                'the_id, the_name, and the_parent must all be string type, got '
                'types {}, {}, {}'.format(type(the_id), type(the_name), type(the_parent)))
        the_id = the_id.strip()
        the_name = the_name.strip()
        the_parent = the_parent.strip()

        # verify that values are permitted and sensible
        if the_id == '':
            raise ValueError('the_id value `` is reserved.')
        if the_name == '':
            raise ValueError('the_name value `` is not permitted.')
        if the_id == the_parent:
            raise ValueError('the_id cannot be the same as the_parent.')

        # try to determine parent from name if not a valid id
        if the_parent != '' and the_parent not in self.labels:
            prospective_parent = self.get_id_from_name(the_parent)
            if prospective_parent is None:
                raise ValueError('the_parent {} matches neither an existing id or name.'.format(the_parent))
            the_parent = prospective_parent

        return the_id, the_name, the_parent

    def add_entry(
            self,
            the_id: str,
            the_name: str,
            the_parent: str = '') -> None:
        """
        Adds a new entry. Note that leading or trailing blanks will be trimmed
        from all input values.

        Parameters
        ----------
        the_id : str
            The id for the label.
        the_name : str
            The name for the label.
        the_parent : str
            The parent id, where blank denotes no parent.

        Returns
        -------
        None
        """

        # validate inputs
        the_id, the_name, the_parent = self._validate_entry(the_id, the_name, the_parent)

        # verify that the_id doesn't already exist
        if the_id in self.labels:
            raise KeyError('the_id = {} already exists'.format(the_id))

        # check if name is already being used, and warn if so
        for key, value in self.labels.items():
            if value == the_name:
                logger.warning(
                    'Note that id {} is already using name {}. Having repeated names is '
                    'permitted, but may lead to confusion.'.format(key, value))

        # add the entry into the labels and subtypes dicts and reset the values
        # perform copy in case of failure
        labels = self.labels.copy()
        subtypes = self.subtypes.copy()
        labels[the_id] = the_name
        if the_parent in subtypes:
            subtypes[the_parent].append(the_id)
        else:
            subtypes[the_parent] = [the_id, ]

        try:
            self.set_labels_and_subtypes(labels, subtypes)
        except (ValueError, KeyError) as e:
            logger.error(
                'Setting new entry id {}, name {}, and parent {} failed with '
                'exception {}'.format(the_id, the_name, the_parent, e))

    def change_entry(
            self,
            the_id: str,
            the_name: str,
            the_parent: str) -> bool:
        """
        Modify the values for a schema element.

        Parameters
        ----------
        the_id : str
        the_name : str
        the_parent : str

        Returns
        -------
        bool
            True if anything was actually changed. False otherwise.
        """

        # validate inputs
        the_id, the_name, the_parent = self._validate_entry(the_id, the_name, the_parent)

        # verify that the_id does already exist
        if the_id not in self.labels:
            raise KeyError('the_id = {} does not exist'.format(the_id))

        # check current values
        current_name = self.labels[the_id]
        current_parents = self.parent_types[the_id]
        current_parent = current_parents[1] if len(current_parents) > 1 else ''

        if current_name == the_name and current_parent == the_parent:
            # nothing is changing
            return False

        # check if name is already being used by a different element, and warn if so
        if current_name != the_name:
            for key, value in self.labels.items():
                if value == the_name and key != the_id:
                    logger.warning(
                        'Note that id {} is already using name {}. Having repeated names is '
                        'permitted, but may lead to confusion.'.format(key, value))

        if current_parent != the_parent:
            labels = self.labels.copy()
            labels[the_id] = the_name
            subtypes = self.subtypes.copy()
            # remove the_id from it's current subtype
            subtypes[current_parent].remove(the_id)
            # add it to the new one
            if the_parent in subtypes:
                subtypes[the_parent].append(the_id)
            else:
                subtypes[the_parent] = [the_id, ]
            try:
                self.set_labels_and_subtypes(labels, subtypes)
            except (ValueError, KeyError) as e:
                logger.error(
                    'Modifying entry id {}, name {}, and parent {} failed with '
                    'exception {}.'.format(the_id, the_name, the_parent, e))
                raise e
        else:
            # just changing the name
            self.labels[the_id] = the_name
        return True

    def delete_entry(
            self,
            the_id: str,
            recursive: bool = False) -> None:
        """
        Deletes the entry from the schema.

        If the given element has children and `recursive=False`, a ValueError
        will be raised. If the given element has children and `recursive=True`,
        then all children will be deleted.

        Parameters
        ----------
        the_id : str
        recursive : bool
        """

        if the_id in self._subtypes:
            # handle all the children
            children = self.subtypes[the_id]
            if children is not None and len(children) > 0:
                if not recursive:
                    raise ValueError(
                        'LabelSchema entry for id {} has children. Either move children to a '
                        'different parent, or make recursive=True to delete all children.'.format(the_id))
                the_children = children.copy()  # unsafe to loop over a changing list
                for entry in the_children:
                    self.delete_entry(entry, recursive=True)
            # now, all the children have been deleted.
            del self._subtypes[the_id]
        # remove the entry from the parent's subtypes list
        parent_id = self.get_parent(the_id)
        self.subtypes[parent_id].remove(parent_id)
        # remove entry from labels
        del self._labels[the_id]
        del self._parent_types[the_id]

    def reorder_child_element(
            self,
            the_id: str,
            spaces: int = 1) -> bool:
        """
        Move the one space (forward or backward) in the list of children for the
        current parent. This is explicitly changes no actual parent/child
        relationships, and only changes the child list ORDERING.

        Parameters
        ----------
        the_id : str
        spaces : int
            How many spaces to shift the entry.

        Returns
        -------
        bool
            True of something actually changed, False otherwise.
        """

        if the_id not in self._labels:
            raise KeyError('No id {}'.format(the_id))
        parent_id = self.get_parent(the_id)
        children = self.subtypes[parent_id]
        # get the current location
        current_index = children.index(the_id)
        # determine the feasible new location
        if spaces < 0:
            new_index = max(0, current_index + spaces)
        else:
            new_index = min(len(children) - 1, current_index + spaces)
        if current_index == new_index:
            return False  # nothing to be done

        # pop our entry out of its current location
        children.pop(current_index)
        # insert it in its new location
        children.insert(new_index, the_id)
        return True

    @classmethod
    def from_file(cls, file_name: str):
        """
        Read schema from a file.

        Parameters
        ----------
        file_name : str

        Returns
        -------
        LabelSchema
        """

        with open(file_name, 'r') as fi:
            input_dict = json.load(fi)
        return cls.from_dict(input_dict)

    @classmethod
    def from_dict(cls, input_dict: Dict):
        """
        Construct from a dictionary.

        Parameters
        ----------
        input_dict : dict

        Returns
        -------
        LabelSchema
        """

        version = input_dict['version']
        labels = input_dict['labels']
        version_date = input_dict.get('version_date', None)
        classification = input_dict.get('classification', 'UNCLASSIFIED')
        subtypes = input_dict.get('subtypes', None)
        conf_values = input_dict.get('confidence_values', None)
        perm_geometries = input_dict.get('permitted_geometries', None)
        return cls(
            version, labels, version_date=version_date, classification=classification,
            subtypes=subtypes, confidence_values=conf_values, permitted_geometries=perm_geometries)

    def to_dict(self) -> Dict:
        """
        Serialize to a dictionary representation.

        Returns
        -------
        dict
        """

        out = OrderedDict()
        out['version'] = self.version
        out['version_date'] = self.version_date
        out['classification'] = self.classification
        if self.confidence_values is not None:
            out['confidence_values'] = self.confidence_values
        if self.permitted_geometries is not None:
            out['permitted_geometries'] = self.permitted_geometries
        out['labels'] = self._labels
        out['subtypes'] = self._subtypes
        return out

    def to_file(self, file_name: str) -> None:
        """
        Write to a (json) file.

        Parameters
        ----------
        file_name : str

        Returns
        -------
        None
        """

        with open(file_name, 'w') as fi:
            json.dump(self.to_dict(), fi, indent=1)

    def is_valid_confidence(self, value: List) -> bool:
        """
        Is the given value a valid confidence (i.e. is in `confidence_values`)?
        Note that `None` is always considered valid here.

        Parameters
        ----------
        value

        Returns
        -------
        bool
        """

        if self._confidence_values is None or value is None:
            return True
        else:
            return value in self._confidence_values

    def is_valid_geometry(self, value: List) -> bool:
        """
        Is the given geometry type allowed (i.e. is in `permitted_geometries`)?
        Note that `None` is always considered valid here.

        Parameters
        ----------
        value : str|Geometry

        Returns
        -------
        bool
        """

        def check_geom(geom):
            if isinstance(geom, (Point, MultiPoint)):
                out = 'point' in self._permitted_geometries
                if not out:
                    logger.error('Not allowed point type geometry components')
                return out
            elif isinstance(geom, (LineString, MultiLineString)):
                out = 'line' in self._permitted_geometries
                if not out:
                    logger.error('Not allowed line type geometry components')
                return out
            elif isinstance(geom, (Polygon, MultiPolygon)):
                out = 'polygon' in self._permitted_geometries
                if not out:
                    logger.error('Not allowed polygon type geometry components')
                return out
            elif isinstance(geom, GeometryCollection):
                out = True
                for entry in geom.geometries:
                    out &= check_geom(entry)
                return out
            else:
                raise TypeError('Got unexpected geometry type `{}`'.format(type(geom)))

        if self._permitted_geometries is None or value is None:
            return True

        if isinstance(value, str):
            return value.lower().strip() in self._permitted_geometries
        if not isinstance(value, Geometry):
            raise TypeError('Got unexpected geometry type `{}`'.format(type(value)))
        return check_geom(value)


##########
# elements for labeling a feature

class LabelMetadata(Jsonable):
    """
    Basic annotation metadata building block - everything but the geometry object
    """

    __slots__ = ('label_id', 'user_id', 'comment', 'confidence', 'timestamp')
    _type = 'LabelMetadata'

    def __init__(
            self,
            label_id: Optional[str] = None,
            user_id: Optional[str] = None,
            comment: Optional[str] = None,
            confidence: Union[None, int, str] = None,
            timestamp: Union[None, int, float] = None):
        """

        Parameters
        ----------
        label_id : None|str
            The label id
        user_id : None|str
            The user id - will default to current user name
        comment : None|str
        confidence : None|str|int
            The confidence value
        timestamp : None|float|int
            The POSIX timestamp (in seconds) - should be construction time.
        """

        self.label_id = label_id  # type: Union[None, str]
        if user_id is None:
            user_id = getpass.getuser()
        self.user_id = user_id  # type: str
        self.comment = comment  # type: Union[None, str]
        self.confidence = confidence  # type: Union[None, str, int]

        if timestamp is None:
            timestamp = time.time()
        if not isinstance(timestamp, float):
            timestamp = float(timestamp)
        self.timestamp = timestamp  # type: float

    @classmethod
    def from_dict(cls, the_json: Dict):
        typ = the_json['type']
        if typ != cls._type:
            raise ValueError('LabelMetadata cannot be constructed from {}'.format(the_json))
        return cls(
            label_id=the_json.get('label_id', None),
            user_id=the_json.get('user_id', None),
            comment=the_json.get('comment', None),
            confidence=the_json.get('confidence', None),
            timestamp=the_json.get('timestamp', None))

    def to_dict(self, parent_dict: Optional[Dict] = None):
        if parent_dict is None:
            parent_dict = OrderedDict()
        parent_dict['type'] = self.type
        for attr in self.__slots__:
            parent_dict[attr] = getattr(self, attr)
        return parent_dict

    def replicate(self):
        kwargs = {}
        for attr in self.__slots__:
            if attr not in ['user_id', 'timestamp']:
                kwargs[attr] = getattr(self, attr)
        the_type = self.__class__
        return the_type(**kwargs)


class LabelMetadataList(Jsonable):
    """
    The collection of LabelMetadata elements.
    """

    __slots__ = ('_elements', )
    _type = 'LabelMetadataList'

    def __init__(self, elements: Union[None, List[LabelMetadata], Dict] = None):
        """

        Parameters
        ----------
        elements : None|List[LabelMetadata|dict]
        """

        self._elements = None
        if elements is not None:
            self.elements = elements

    def __len__(self):
        if self._elements is None:
            return 0
        return len(self._elements)

    def __getitem__(self, item):
        # type: (Any) -> LabelMetadata
        if self._elements is None:
            raise StopIteration
        return self._elements[item]

    @property
    def elements(self) -> Optional[List[LabelMetadata]]:
        """
        The LabelMetadata elements.

        Returns
        -------
        None|List[LabelMetadata]
        """

        return self._elements

    @elements.setter
    def elements(self, elements):
        if elements is None:
            self._elements = None
        if not isinstance(elements, list):
            raise TypeError('elements must be a list of LabelMetadata elements')
        self._elements = []
        for element in elements:
            self.insert_new_element(element)

    def insert_new_element(self, element: LabelMetadata) -> None:
        """
        Inserts an element at the head of the elements list.

        Parameters
        ----------
        element : LabelMetadata

        Returns
        -------
        None
        """

        if isinstance(element, dict):
            element = LabelMetadata.from_dict(element)
        if not isinstance(element, LabelMetadata):
            raise TypeError('element must be an LabelMetadata instance, got type {}'.format(type(element)))

        if self._elements is None:
            self._elements = [element, ]
        elif len(self._elements) == 0:
            self._elements.append(element)
        else:
            for i, entry in enumerate(self._elements):
                if element.timestamp > entry.timestamp:
                    self._elements.insert(i, element)
                    break

    @classmethod
    def from_dict(cls, the_json):
        # type: (dict) -> LabelMetadataList
        typ = the_json['type']
        if typ != cls._type:
            raise ValueError('LabelMetadataList cannot be constructed from {}'.format(the_json))
        return cls(elements=the_json.get('elements', None))

    def to_dict(self, parent_dict: Optional[Dict] = None):
        if parent_dict is None:
            parent_dict = OrderedDict()
        parent_dict['type'] = self.type
        if self._elements is None:
            parent_dict['elements'] = None
        else:
            parent_dict['elements'] = [entry.to_dict() for entry in self._elements]
        return parent_dict

    def replicate(self):  # type: () ->  LabelMetadataList
        kwargs = {}
        elements = self.elements
        if elements is not None:
            kwargs['elements'] = [elements[0].replicate()]
        the_type = self.__class__
        return the_type(**kwargs)

    def get_label_id(self) -> Optional[str]:
        """
        Gets the current label id.

        Returns
        -------
        None|str
        """

        return None if (self.elements is None or len(self.elements) == 0) else self.elements[0].label_id


class LabelProperties(AnnotationProperties):
    _type = 'LabelProperties'

    @property
    def parameters(self):
        """
        LabelMetadataList: The parameters
        """

        return self._parameters

    @parameters.setter
    def parameters(self, value):
        if value is None:
            self._parameters = LabelMetadataList()
            return
        if isinstance(value, dict):
            self._parameters = LabelMetadataList.from_dict(value)
            return
        if isinstance(value, LabelMetadataList):
            self._parameters = value
            return
        raise TypeError('Got unexpected type for parameters `{}`'.format(type(value)))

    def get_label_id(self):
        """
        Gets the current label id.

        Returns
        -------
        None|str
        """

        return None if self.parameters is None else self.parameters.get_label_id()


############
# the feature extensions

class LabelFeature(AnnotationFeature):
    """
    A specific extension of the Feature class which has the properties attribute
    populated with LabelProperties instance.
    """

    @property
    def properties(self):
        """
        The properties.

        Returns
        -------
        None|LabelProperties
        """

        return self._properties

    @properties.setter
    def properties(self, properties):
        if properties is None:
            self._properties = LabelProperties()
            return
        if isinstance(properties, dict):
            self._properties = LabelProperties.from_dict(properties)
            return
        if isinstance(properties, LabelProperties):
            self._properties = properties
            return
        raise TypeError('properties must be an LabelProperties')

    def add_annotation_metadata(self, value):
        """
        Adds the new label to the series of labeling efforts.

        Parameters
        ----------
        value : LabelMetadata
        """

        if self._properties is None:
            self._properties = LabelProperties()
        self._properties.parameters.insert_new_element(value)

    def get_label_id(self):
        """
        Gets the label id.

        Returns
        -------
        None|str
        """

        return None if self.properties is None else self.properties.get_label_id()


class LabelCollection(AnnotationCollection):
    """
    A specific extension of the FeatureCollection class which has the features are
    LabelFeature instances.
    """

    @property
    def features(self):
        """
        The features list.

        Returns
        -------
        List[LabelFeature]
        """

        return self._features

    @features.setter
    def features(self, features):
        if features is None:
            self._features = None
            self._feature_dict = None
            return

        if not isinstance(features, list):
            raise TypeError(
                'features must be a list of LabelFeatures. '
                'Got {}'.format(type(features)))

        for entry in features:
            self.add_feature(entry)

    def add_feature(self, feature):
        """
        Add an annotation.

        Parameters
        ----------
        feature : LabelFeature
        """

        if isinstance(feature, dict):
            feature = LabelFeature.from_dict(feature)
        if not isinstance(feature, LabelFeature):
            raise TypeError('This requires an LabelFeature instance, got {}'.format(type(feature)))

        if self._features is None:
            self._feature_dict = {feature.uid: 0}
            self._features = [feature, ]
        else:
            self._feature_dict[feature.uid] = len(self._features)
            self._features.append(feature)

    def __getitem__(self, item):
        # type: (Any) -> Union[LabelFeature, List[LabelFeature]]
        if self._features is None:
            raise StopIteration

        if isinstance(item, str):
            index = self._feature_dict[item]
            return self._features[index]
        return self._features[item]


###########
# serialized file object

class FileLabelCollection(FileAnnotationCollection):
    """
    An collection of annotation elements associated with a given single image element file.
    """

    __slots__ = (
        '_version', '_label_schema', '_image_file_name', '_image_id', '_core_name', '_annotations')
    _type = 'FileLabelCollection'

    def __init__(self, label_schema, version=None, annotations=None,
                 image_file_name=None, image_id=None, core_name=None):
        if version is None:
            version = _LABEL_VERSION

        if isinstance(label_schema, str):
            label_schema = LabelSchema.from_file(label_schema)
        elif isinstance(label_schema, dict):
            label_schema = LabelSchema.from_dict(label_schema)
        if not isinstance(label_schema, LabelSchema):
            raise TypeError('label_schema must be an instance of a LabelSchema.')
        self._label_schema = label_schema

        FileAnnotationCollection.__init__(
            self, version=version, annotations=annotations, image_file_name=image_file_name,
            image_id=image_id, core_name=core_name)

    @property
    def label_schema(self):
        """
        The label schema.

        Returns
        -------
        LabelSchema
        """
        return self._label_schema

    @property
    def annotations(self):
        """
        The annotations.

        Returns
        -------
        LabelCollection
        """

        return self._annotations

    @annotations.setter
    def annotations(self, annotations):
        # type: (Union[None, LabelCollection, dict]) -> None
        if annotations is None:
            self._annotations = None
            return

        if isinstance(annotations, LabelCollection):
            self._annotations = annotations
        elif isinstance(annotations, dict):
            self._annotations = LabelCollection.from_dict(annotations)
        else:
            raise TypeError(
                'annotations must be an LabelCollection. Got type {}'.format(type(annotations)))
        self.validate_annotations(strict=False)

    def add_annotation(self, annotation, validate_confidence=True, validate_geometry=True):
        """
        Add an annotation, with a check for valid values in confidence and geometry type.

        Parameters
        ----------
        annotation : LabelFeature
            The prospective annotation.
        validate_confidence : bool
            Should we check that all confidence values follow the schema?
        validate_geometry : bool
            Should we check that all geometries are of allowed type?

        Returns
        -------
        None
        """

        if not isinstance(annotation, LabelFeature):
            raise TypeError('This requires an LabelFeature instance. Got {}'.format(type(annotation)))

        if self._annotations is None:
            self._annotations = LabelCollection()

        valid = True
        if validate_confidence:
            valid &= self._valid_confidences(annotation)
        if validate_geometry:
            valid &= self._valid_geometry(annotation)
        if not valid:
            raise ValueError('LabelFeature does not follow the schema.')
        self._annotations.add_feature(annotation)

    def is_annotation_valid(self, annotation):
        """
        Is the given annotation valid according to the schema?

        Parameters
        ----------
        annotation : LabelFeature

        Returns
        -------
        bool
        """

        if not isinstance(annotation, LabelFeature):
            return False

        if self._label_schema is None:
            return True
        valid = self._valid_confidences(annotation)
        valid &= self._valid_geometry(annotation)
        return valid

    def _valid_confidences(self, annotation):
        if self._label_schema is None:
            return True

        if annotation.properties is None or annotation.properties.parameters is None:
            return True

        valid = True
        for entry in annotation.properties.parameters:
            if not self._label_schema.is_valid_confidence(entry.confidence):
                valid = False
                logger.error('Invalid confidence value {}'.format(entry.confidence))
        return valid

    def _valid_geometry(self, annotation):
        if self._label_schema is None:
            return True
        if not self._label_schema.is_valid_geometry(annotation.geometry):
            logger.error('Invalid geometry type {}'.format(type(annotation.geometry)))
            return False
        return True

    def validate_annotations(self, strict=True):
        if self._annotations is None:
            return True

        valid = True
        for entry in self._annotations:
            valid &= self.is_annotation_valid(entry)
        if strict and not valid:
            raise ValueError('Some annotation does not follow the schema.')
        return valid

    @classmethod
    def from_file(cls, file_name):
        """
        Read from (json) file.

        Parameters
        ----------
        file_name : str

        Returns
        -------
        FileLabelCollection
        """

        with open(file_name, 'r') as fi:
            the_dict = json.load(fi)
        return cls.from_dict(the_dict)

    @classmethod
    def from_dict(cls, the_dict):
        """
        Define from a dictionary representation.

        Parameters
        ----------
        the_dict : dict

        Returns
        -------
        FileLabelCollection
        """

        if not isinstance(the_dict, dict):
            raise TypeError('This requires a dict. Got type {}'.format(type(the_dict)))
        if 'label_schema' not in the_dict:
            raise KeyError('this dictionary must contain a label_schema')

        typ = the_dict.get('type', 'NONE')
        if typ != cls._type:
            raise ValueError('FileLabelCollection cannot be constructed from the input dictionary')

        return cls(
            the_dict['label_schema'],
            version=the_dict.get('version', 'UNKNOWN'),
            annotations=the_dict.get('annotations', None),
            image_file_name=the_dict.get('image_file_name', None),
            image_id=the_dict.get('image_id', None),
            core_name=the_dict.get('core_name', None))

    def to_dict(self, parent_dict=None):
        if parent_dict is None:
            parent_dict = OrderedDict()
        parent_dict['type'] = self.type
        parent_dict['version'] = self.version
        parent_dict['label_schema'] = self.label_schema.to_dict()
        if self.image_file_name is not None:
            parent_dict['image_file_name'] = self.image_file_name
        if self.image_id is not None:
            parent_dict['image_id'] = self.image_id
        if self.core_name is not None:
            parent_dict['core_name'] = self.core_name
        if self.annotations is not None:
            parent_dict['annotations'] = self.annotations.to_dict()
        return parent_dict



from .__about__ import *
import logging


__all__ = ['__version__',
           '__classification__', '__author__', '__url__', '__email__',
           '__title__', '__summary__',
           '__license__', '__copyright__']


# establish logging paradigm
logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class PT(TREElement):
    def __init__(self, value):
        super(PT, self).__init__()
        self.add_field('LON', 's', 15, value)
        self.add_field('LAT', 's', 15, value)


class BP(TREElement):
    def __init__(self, value):
        super(BP, self).__init__()
        self.add_field('NUM_PTS', 'd', 2, value)
        self.add_loop('PTs', self.NUM_PTS, PT, value)


class BND(TREElement):
    def __init__(self, value):
        super(BND, self).__init__()
        self.add_field('BID', 's', 5, value)
        self.add_field('WS1', 's', 5, value)
        self.add_field('WS2', 's', 5, value)


class AUX(TREElement):
    def __init__(self, value):
        super(AUX, self).__init__()
        self.add_field('API', 's', 20, value)
        self.add_field('APF', 's', 1, value)
        self.add_field('UNIAPX', 's', 7, value)
        self.add_field('APN', 's', 10, value)
        self.add_field('APR', 's', 20, value)
        self.add_field('APA', 's', 20, value)


class SNS(TREElement):
    def __init__(self, value):
        super(SNS, self).__init__()
        self.add_field('NUM_BP', 'd', 2, value)
        self.add_loop('BPs', self.NUM_BP, BP, value)
        self.add_field('NUM_BND', 'd', 2, value)
        self.add_loop('BNDs', self.NUM_BND, BND, value)
        self.add_field('UNIRES', 's', 3, value)
        self.add_field('REX', 's', 6, value)
        self.add_field('REY', 's', 6, value)
        self.add_field('GSX', 's', 6, value)
        self.add_field('GSY', 's', 6, value)
        self.add_field('GSL', 's', 12, value)
        self.add_field('PLTFM', 's', 8, value)
        self.add_field('INS', 's', 8, value)
        self.add_field('MOD', 's', 4, value)
        self.add_field('PRL', 's', 5, value)
        self.add_field('ACT', 's', 18, value)
        self.add_field('UNINOA', 's', 3, value)
        self.add_field('NOA', 's', 7, value)
        self.add_field('UNIANG', 's', 3, value)
        self.add_field('ANG', 's', 7, value)
        self.add_field('UNIALT', 's', 3, value)
        self.add_field('ALT', 's', 9, value)
        self.add_field('LONSCC', 's', 10, value)
        self.add_field('LATSCC', 's', 10, value)
        self.add_field('UNISAE', 's', 3, value)
        self.add_field('SAZ', 's', 7, value)
        self.add_field('SEL', 's', 7, value)
        self.add_field('UNIRPY', 's', 3, value)
        self.add_field('ROL', 's', 7, value)
        self.add_field('PIT', 's', 7, value)
        self.add_field('YAW', 's', 7, value)
        self.add_field('UNIPXT', 's', 3, value)
        self.add_field('PIXT', 's', 14, value)
        self.add_field('UNISPE', 's', 7, value)
        self.add_field('ROS', 's', 22, value)
        self.add_field('PIS', 's', 22, value)
        self.add_field('YAS', 's', 22, value)
        self.add_field('NUM_AUX', 'd', 3, value)
        self.add_loop('AUXs', self.NUM_AUX, AUX, value)


class SNSPSBType(TREElement):
    def __init__(self, value):
        super(SNSPSBType, self).__init__()
        self.add_field('NUMSNS', 'd', 2, value)
        self.add_loop('SNSs', self.NUMSNS, SNS, value)


class SNSPSB(TREExtension):
    _tag_value = 'SNSPSB'
    _data_type = SNSPSBType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class MAPLOBType(TREElement):
    def __init__(self, value):
        super(MAPLOBType, self).__init__()
        self.add_field('UNILOA', 's', 3, value)
        self.add_field('LOD', 's', 5, value)
        self.add_field('LAD', 's', 5, value)
        self.add_field('LSO', 's', 15, value)
        self.add_field('PSO', 's', 15, value)


class MAPLOB(TREExtension):
    _tag_value = 'MAPLOB'
    _data_type = MAPLOBType




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class ACFTA_132Type(TREElement):
    def __init__(self, value):
        super(ACFTA_132Type, self).__init__()
        self.add_field('AC_MSN_ID', 's', 10, value)
        self.add_field('SCTYPE', 's', 1, value)
        self.add_field('SCNUM', 's', 4, value)
        self.add_field('SENSOR_ID', 's', 3, value)
        self.add_field('PATCH_TOT', 's', 4, value)
        self.add_field('MTI_TOT', 's', 3, value)
        self.add_field('PDATE', 's', 7, value)
        self.add_field('IMHOSTNO', 's', 3, value)
        self.add_field('IMREQID', 's', 5, value)
        self.add_field('SCENE_SOURCE', 's', 1, value)
        self.add_field('MPLAN', 's', 2, value)
        self.add_field('ENTLOC', 's', 21, value)
        self.add_field('ENTALT', 's', 6, value)
        self.add_field('EXITLOC', 's', 21, value)
        self.add_field('EXITALT', 's', 6, value)
        self.add_field('TMAP', 's', 7, value)
        self.add_field('RCS', 's', 3, value)
        self.add_field('ROW_SPACING', 's', 7, value)
        self.add_field('COL_SPACING', 's', 7, value)
        self.add_field('SENSERIAL', 's', 4, value)
        self.add_field('ABSWVER', 's', 7, value)


class ACFTA_132(TREExtension):
    _tag_value = 'ACFTA'
    _data_type = ACFTA_132Type


class ACFTA_154Type(TREElement):
    def __init__(self, value):
        super(ACFTA_154Type, self).__init__()
        self.add_field('AC_MSN_ID', 's', 10, value)
        self.add_field('AC_TAIL_NO', 's', 10, value)
        self.add_field('SENSOR_ID', 's', 10, value)
        self.add_field('SCENE_SOURCE', 's', 1, value)
        self.add_field('SCNUM', 's', 6, value)
        self.add_field('PDATE', 's', 8, value)
        self.add_field('IMHOSTNO', 's', 6, value)
        self.add_field('IMREQID', 's', 5, value)
        self.add_field('MPLAN', 's', 3, value)
        self.add_field('ENTLOC', 's', 21, value)
        self.add_field('ENTALT', 's', 6, value)
        self.add_field('EXITLOC', 's', 21, value)
        self.add_field('EXITALT', 's', 6, value)
        self.add_field('TMAP', 's', 7, value)
        self.add_field('ROW_SPACING', 's', 7, value)
        self.add_field('COL_SPACING', 's', 7, value)
        self.add_field('SENSERIAL', 's', 6, value)
        self.add_field('ABSWVER', 's', 7, value)
        self.add_field('PATCH_TOT', 's', 4, value)
        self.add_field('MTI_TOT', 's', 3, value)


class ACFTA_154(TREExtension):
    _tag_value = 'ACFTA'
    _data_type = ACFTA_154Type


class ACFTA_199Type(TREElement):
    def __init__(self, value):
        super(ACFTA_199Type, self).__init__()
        self.add_field('AC_MSN_ID', 's', 20, value)
        self.add_field('AC_TAIL_NO', 's', 10, value)
        self.add_field('AC_TO', 's', 12, value)
        self.add_field('SENSOR_ID_TYPE', 's', 4, value)
        self.add_field('SENSOR_ID', 's', 6, value)
        self.add_field('SCENE_SOURCE', 's', 1, value)
        self.add_field('SCNUM', 's', 6, value)
        self.add_field('PDATE', 's', 8, value)
        self.add_field('IMHOSTNO', 's', 6, value)
        self.add_field('IMREQID', 's', 5, value)
        self.add_field('MPLAN', 's', 3, value)
        self.add_field('ENTLOC', 's', 25, value)
        self.add_field('ENTELV', 's', 6, value)
        self.add_field('ELVUNIT', 's', 1, value)
        self.add_field('EXITLOC', 's', 25, value)
        self.add_field('EXITELV', 's', 6, value)
        self.add_field('TMAP', 's', 7, value)
        self.add_field('RESERVD1', 's', 1, value)
        self.add_field('ROW_SPACING', 's', 7, value)
        self.add_field('COL_SPACING', 's', 7, value)
        self.add_field('FOCAL_LENGTH', 's', 6, value)
        self.add_field('SENSERIAL', 's', 6, value)
        self.add_field('ABSWVER', 's', 7, value)
        self.add_field('CAL_DATE', 's', 8, value)
        self.add_field('PATCH_TOT', 's', 4, value)
        self.add_field('MTI_TOT', 's', 3, value)


class ACFTA_199(TREExtension):
    _tag_value = 'ACFTA'
    _data_type = ACFTA_199Type


class ACFTA(TREExtension):
    _tag_value = 'ACFTA'

    def __init__(self):
        raise ValueError(
            'Not to be implemented directly. '
            'Use of one ACFTA_132, ACFTA_154, or ACFTA_199')

    @classmethod
    def from_bytes(cls, value, start):
        """

        Parameters
        ----------
        value : bytes
        start : int

        Returns
        -------
        ACFTA_132|ACFTA_154|ACFTA_199
        """

        tag_value = value[start:start+6].decode('utf-8').strip()
        if tag_value != cls._tag_value:
            raise ValueError('tag value must be {}. Got {}'.format(cls._tag_value, tag_value))

        lng = int(value[start+6:start+11])
        if lng == 132:
            return ACFTA_132.from_bytes(value, start)
        elif lng == 154:
            return ACFTA_154.from_bytes(value, start)
        elif lng == 199:
            return ACFTA_199.from_bytes(value, start)
        else:
            raise ValueError('the data must be length 132, 154, or 199. Got {}'.format(lng))



#
# Copyright 2020-2021 Valkyrie Systems Corporation
#
# Licensed under MIT License.  See LICENSE.
#

__classification__ = "UNCLASSIFIED"
__author__ = "Nathan Bombaci, Valkyrie"


from typing import List
import numpy as np


def parse_text(elem):
    """
    Reverse of `xml.make_elem` by converting an element's text string to
    an int, float, bool, or str, as appropriate.

    Parameters
    ----------
    elem: lxml.etree.ElementTree.Element
        Element to convert to the most restrictive python type possible.

    Returns
    -------
    val: int|float|bool|str
        Converted value.
    """

    for converter in (int, float, parse_bool_text, str):
        try:
            val = converter(elem.text)
            break
        except ValueError:
            continue
    return val


def parse_bool_text(text):
    """
    Gets a boolean from a string.

    Parameters
    ----------
    text: str
        One of `'true', '1', 'false', '0'`.

    Returns
    -------
    val: bool
        Boolean value converted from `text`.

    Raises
    ------
    ValueError
        The text string is not either ``'true'`` or ``'false'``.
    """

    text = text.lower()
    if text in ['true', '1']:
        return True
    if text in ['false', '0']:
        return False
    raise ValueError("Cannot parse bool from {}".format(text))


def parse_bool(elem):
    """
    Gets a boolean from an element.

    Parameters
    ----------
    elem : lxml.etree.ElementTree.Element
        Element to convert.

    Returns
    -------
    val : bool
        Boolean value of the `elem`'s text.
    """

    return parse_bool_text(elem.text)


def parse_sequence(node, keys, conversion=parse_text):
    """
    Reverse of `sequence_node`.

    Parameters
    ----------
    node : lxml.etree.ElementTree.Element
        Element containing a sequence node.
    keys : List
        List of element names to parse.
    conversion : Callable
        Conversion function. (Default: `parse_text`)

    Returns
    ------
    List
        List of parsed values, one for each element of `keys`.
    """

    return [conversion(node.find('./' + key)) for key in keys]


def parse_xyz(node):
    """
    Parse a node with ``'X'``, ``'Y'``, and ``'Z'`` children

    Parameters
    ----------
    node : lxml.etree.ElementTree.Element
        Element containing an XYZ sequence node.

    Returns
    -------
    List
        List [X, Y, Z]. Parsed values.
    """

    return parse_sequence(node, ['X', 'Y', 'Z'], lambda x: float(x.text))


def parse_xy(node):
    """
    Parse a node with ``'X'`` and ``'Y'`` children.

    Parameters
    ----------
    node: lxml.etree.ElementTree.Element
        Element containing an XY sequence node.

    Returns
    -------
    List
        List [X, Y]. Parsed values.
    """

    return parse_sequence(node, ['X', 'Y'], lambda x: float(x.text))


def parse_ll(node):
    """
    Parse a node with ``'Lat'`` and ``'Lon'`` children.

    Parameters
    ----------
    node: lxml.etree.ElementTree.Element
        Element containing a Lat/Lon sequence node.

    Returns
    -------
    List
        List [Lon, Lat]. Parsed values as radians.
    """

    return np.radians(float(node.findtext('Lon'))), np.radians(float(node.findtext('Lat')))


def parse_llh(node):
    """
    Parse a node with ``'Lat'``, ``'Lon'``, ``'HAE'`` children.
    Parameters
    ----------
    node: lxml.etree.ElementTree.Element
        Element containing a Lat/Lon/HAE sequence node.
    Returns
    -------
    List
        List [Lon, Lat, HAE]. Parsed Lon, Lat values as radians, HAE value as meters.
    """
    return np.radians(float(node.findtext('Lon'))),\
        np.radians(float(node.findtext('Lat'))),\
        float(node.findtext('HAE'))


def parse_poly2d(node):
    """
    Parse a node with ``'exponent1'`` and ``'exponent2'`` children.

    Args
    ----
    node: `lxml.etree.ElementTree.Element`
        Element containing a poly2d node.

    Returns
    -------
    result: list of list, shape=(:, :)
        A list of coefficient values.

    """
    coefs = node.findall('./Coef')
    num_coefs1 = max([int(coef.get('exponent1')) for coef in coefs]) + 1
    num_coefs2 = max([int(coef.get('exponent2')) for coef in coefs]) + 1
    poly2d = np.zeros((num_coefs1, num_coefs2), np.float64)

    for coef in coefs:
        poly2d[int(coef.get('exponent1')), int(coef.get('exponent2'))] = float(coef.text)

    return poly2d.tolist()



"""
A utility for dumping a NITF header to the console. Contributed by Austin Lan of L3/Harris.

To dump NITF header information to a text file from the command-line

>>> python -m sarpy.utils.nitf_utils <path to nitf file>

For a basic help on the command-line, check

>>> python -m sarpy.utils.nitf_utils --help

"""

__classification__ = "UNCLASSIFIED"
__author__ = "Austin Lan, L3/Harris"


import argparse
import functools
import sys
from xml.dom import minidom
import os
from typing import Union, BinaryIO, TextIO, List, Dict
from io import StringIO

from sarpy.io.general.nitf import NITFDetails
from sarpy.io.general.nitf_elements.base import NITFElement, TRE, TREList, UserHeaderType
from sarpy.io.general.nitf_elements.des import DataExtensionHeader, DataExtensionHeader0, \
    DESUserHeader
from sarpy.io.general.nitf_elements.graphics import GraphicsSegmentHeader
from sarpy.io.general.nitf_elements.image import ImageSegmentHeader, ImageSegmentHeader0, MaskSubheader
from sarpy.io.general.nitf_elements.label import LabelSegmentHeader
from sarpy.io.general.nitf_elements.nitf_head import NITFHeader, NITFHeader0
from sarpy.io.general.nitf_elements.res import ReservedExtensionHeader, ReservedExtensionHeader0, \
    RESUserHeader
from sarpy.io.general.nitf_elements.symbol import SymbolSegmentHeader
from sarpy.io.general.nitf_elements.text import TextSegmentHeader, TextSegmentHeader0
from sarpy.io.general.nitf_elements.tres.tre_elements import TREElement


# Custom print function
print_func = print


############
# helper methods

def _filter_files(input_path):
    """
    Determine if a given input path corresponds to a NITF 2.1 or 2.0 file.

    Parameters
    ----------
    input_path : str

    Returns
    -------
    bool
    """

    if not os.path.isfile(input_path):
        return False
    _, fext = os.path.splitext(input_path)
    with open(input_path, 'rb') as fi:
        check = fi.read(9)
    return check in [b'NITF02.10', b'NITF02.00']


def _create_default_output_file(input_file, output_directory=None):
    if not isinstance(input_file, str):
        if output_directory is None:
            return os.path.expanduser('~/Desktop/header_dump.txt')
        else:
            return os.path.join(output_directory, 'header_dump.txt')

    if output_directory is None:
        return os.path.splitext(input_file)[0] + '.header_dump.txt'
    else:
        return os.path.join(output_directory, os.path.splitext(os.path.split(input_file)[1])[0] + '.header_dump.txt')


def _decode_effort(value):
    # type: (bytes) -> Union[bytes, str]

    # noinspection PyBroadException
    try:
        return value.decode()
    except Exception:
        return value


############
# printing methods

def _print_element_field(elem, field, prefix=''):
    # type: (Union[None, NITFElement], Union[None, str], str) -> None
    if elem is None or field is None:
        return

    value = getattr(elem, field, None)
    if value is None:
        value = ''
    print_func('{}{} = {}'.format(prefix, field, value))


def _print_element(elem, prefix=''):
    # type: (Union[None, NITFElement], str) -> None
    if elem is None:
        return

    # noinspection PyProtectedMember
    for field in elem._ordering:
        _print_element_field(elem, field, prefix=prefix)


def _print_element_list(elem_list, prefix=''):
    # type: (Union[None, List[NITFElement]], str) -> None
    if elem_list is None:
        return

    for i, elem in enumerate(elem_list):
        _print_element(elem, prefix='{}[{}].'.format(prefix, i))


def _print_tre_element(field, value, prefix=''):
    # type: (Union[None, str], Union[str, int, bytes], str) -> None
    if field is None:
        return

    if value is None:
        value = ''
    print_func('{}{} = {}'.format(prefix, field, value))


def _print_tre_list(elem_list, prefix=''):
    # type: (Union[None, List, TREList], str) -> None

    if elem_list is None:
        return

    for i, elem in enumerate(elem_list):
        _print_tre_dict(elem, '{}[{}].'.format(prefix, i))


def _print_tre_dict(elem_dict, prefix=''):
    # type: (Union[None, Dict], str) -> None
    if elem_dict is None:
        return

    for field, value in elem_dict.items():
        if isinstance(value, list):
            _print_tre_list(value, '{}{}'.format(prefix, field))
        else:
            _print_tre_element(field, value, prefix)


def _print_tres(tres):
    # type: (Union[TREList, List[TRE]]) -> None
    for tre in tres:
        print_func('')
        if isinstance(tre.DATA, TREElement):
            _print_tre_dict(tre.DATA.to_dict(), prefix='{}.'.format(tre.TAG))
        else:
            # Unknown TRE
            _print_tre_element('DATA', _decode_effort(tre.DATA), prefix='{}.'.format(tre.TAG))


def _print_file_header(hdr):
    # type: (Union[NITFHeader, NITFHeader0]) -> None

    # noinspection PyProtectedMember
    for field in hdr._ordering:
        if field == 'Security':
            _print_element(getattr(hdr, field, None), prefix='FS')
        elif field == 'FBKGC':
            value = getattr(hdr, field, None)
            print_func('FBKGC = {} {} {}'.format(value[0], value[1], value[2]))
        elif field in [
            'ImageSegments', 'GraphicsSegments', 'SymbolSegments', 'LabelSegments',
            'TextSegments', 'DataExtensions', 'ReservedExtensions']:
            pass
        elif field in ['UserHeader', 'ExtendedHeader']:
            value = getattr(hdr, field, None)
            assert(isinstance(value, UserHeaderType))
            if value and value.data and value.data.tres:
                _print_tres(value.data.tres)
        else:
            _print_element_field(hdr, field)


def _print_mask_header(hdr):
    # type: (Union[None, MaskSubheader]) -> None
    if hdr is None:
        return

    print_func('----- Mask Subheader (part of image data segment) -----')
    # noinspection PyProtectedMember
    for field in hdr._ordering:
        if field in ['BMR', 'TMR']:
            value = getattr(hdr, field, None)
            if value is None:
                continue
            else:
                for the_band, subarray in enumerate(value):
                    print_func('{}BND{} = {}'.format(field, the_band, subarray))
        else:
            _print_element_field(hdr, field, prefix='')


def _print_image_header(hdr):
    # type: (Union[ImageSegmentHeader, ImageSegmentHeader0]) -> None

    # noinspection PyProtectedMember
    for field in hdr._ordering:
        if field == 'Security':
            _print_element(getattr(hdr, field, None), prefix='IS')
        elif field in ['Comments', 'Bands']:
            _print_element_list(getattr(hdr, field, None), prefix='{}'.format(field))
        elif field in ['UserHeader', 'ExtendedHeader']:
            value = getattr(hdr, field, None)
            assert(isinstance(value, UserHeaderType))
            if value and value.data and value.data.tres:
                _print_tres(value.data.tres)
        else:
            _print_element_field(hdr, field)
    _print_mask_header(hdr.mask_subheader)


def _print_basic_header(hdr, prefix):
    # noinspection PyProtectedMember
    for field in hdr._ordering:
        if field == 'Security':
            _print_element(getattr(hdr, field, None), prefix=prefix)
        elif field in ['UserHeader', 'ExtendedHeader']:
            value = getattr(hdr, field, None)
            assert(isinstance(value, UserHeaderType))
            if value and value.data and value.data.tres:
                _print_tres(value.data.tres)
        else:
            _print_element_field(hdr, field)


def _print_graphics_header(hdr):
    # type: (GraphicsSegmentHeader) -> None
    _print_basic_header(hdr, 'SS')


def _print_symbol_header(hdr):
    # type: (SymbolSegmentHeader) -> None
    _print_basic_header(hdr, 'SS')


def _print_label_header(hdr):
    # type: (LabelSegmentHeader) -> None
    _print_basic_header(hdr, 'LS')


def _print_text_header(hdr):
    # type: (Union[TextSegmentHeader, TextSegmentHeader0]) -> None
    _print_basic_header(hdr, 'TS')


def _print_extension_header(hdr, prefix):
    # noinspection PyProtectedMember
    for field in hdr._ordering:
        if field == 'Security':
            _print_element(getattr(hdr, field, None), prefix=prefix)
        elif field in ['UserHeader', 'ExtendedHeader']:
            value = getattr(hdr, field, None)
            if isinstance(value, (DESUserHeader, RESUserHeader)):
                if value.data:
                    # Unknown user-defined subheader
                    print_func('{}SHF = {}'.format(prefix, _decode_effort(value.data)))
            else:
                # e.g., XMLDESSubheader
                _print_element(value, prefix='{}SHF.'.format(prefix))
        else:
            _print_element_field(hdr, field)


def _print_des_header(hdr):
    # type: (Union[DataExtensionHeader, DataExtensionHeader0]) -> None
    _print_extension_header(hdr, 'DES')


def _print_res_header(hdr):
    # type: (Union[ReservedExtensionHeader, ReservedExtensionHeader0]) -> None
    _print_extension_header(hdr, 'RES')


def print_nitf(file_name, dest=sys.stdout):
    """
    Worker function to dump the NITF header and various subheader details to the
    provided destination.

    Parameters
    ----------
    file_name : str|BinaryIO
    dest : TextIO
    """

    # Configure print function for desired destination
    #    - e.g., stdout, string buffer, file
    global print_func
    print_func = functools.partial(print, file=dest)

    details = NITFDetails(file_name)

    if isinstance(file_name, str):
        print_func('')
        print_func('Details for file {}'.format(file_name))
        print_func('')

    print_func('----- File Header -----')
    _print_file_header(details.nitf_header)
    print_func('')

    if details.img_subheader_offsets is not None:
        for img_subhead_num in range(details.img_subheader_offsets.size):
            print_func('----- Image {} -----'.format(img_subhead_num))
            hdr = details.parse_image_subheader(img_subhead_num)
            _print_image_header(hdr)
            print_func('')

    if details.graphics_subheader_offsets is not None:
        for graphics_subhead_num in range(details.graphics_subheader_offsets.size):
            print_func('----- Graphic {} -----'.format(graphics_subhead_num))
            hdr = details.parse_graphics_subheader(graphics_subhead_num)
            _print_graphics_header(hdr)
            data = details.get_graphics_bytes(graphics_subhead_num)
            print_func('GSDATA = {}'.format(_decode_effort(data)))
            print_func('')

    if details.symbol_subheader_offsets is not None:
        for symbol_subhead_num in range(details.symbol_subheader_offsets.size):
            print_func('----- Symbol {} -----'.format(symbol_subhead_num))
            hdr = details.parse_symbol_subheader(symbol_subhead_num)
            _print_symbol_header(hdr)
            data = details.get_symbol_bytes(symbol_subhead_num)
            print_func('SSDATA = {}'.format(_decode_effort(data)))
            print_func('')

    if details.label_subheader_offsets is not None:
        for label_subhead_num in range(details.label_subheader_offsets.size):
            print_func('----- Label {} -----'.format(label_subhead_num))
            hdr = details.parse_label_subheader(label_subhead_num)
            _print_label_header(hdr)
            data = details.get_label_bytes(label_subhead_num)
            print_func('LSDATA = {}'.format(_decode_effort(data)))
            print_func('')

    if details.text_subheader_offsets is not None:
        for text_subhead_num in range(details.text_subheader_offsets.size):
            print_func('----- Text {} -----'.format(text_subhead_num))
            hdr = details.parse_text_subheader(text_subhead_num)
            _print_text_header(hdr)
            data = details.get_text_bytes(text_subhead_num)
            print_func('TSDATA = {}'.format(_decode_effort(data)))
            print_func('')

    if details.des_subheader_offsets is not None:
        for des_subhead_num in range(details.des_subheader_offsets.size):
            print_func('----- DES {} -----'.format(des_subhead_num))
            hdr = details.parse_des_subheader(des_subhead_num)
            _print_des_header(hdr)
            data = details.get_des_bytes(des_subhead_num)

            des_id = hdr.DESID if details.nitf_version == '02.10' else hdr.DESTAG

            if des_id.strip() in ['XML_DATA_CONTENT', 'SICD_XML', 'SIDD_XML']:
                xml_str = minidom.parseString(
                    data.decode()).toprettyxml(indent='    ', newl='\n')
                # NB: this may or not exhibit platform dependent choices in which codec (i.e. latin-1 versus utf-8)
                print_func('DESDATA =')
                for line_num, xml_entry in enumerate(xml_str.splitlines()):
                    if line_num == 0:
                        # Remove xml that gets inserted by minidom, if it's not actually there
                        if (not data.startswith(b'<?xml version')) and xml_entry.startswith('<?xml version'):
                            continue
                        print_func(xml_entry)
                    elif xml_entry.strip() != '':
                        # Remove extra new lines if XML is already formatted
                        print_func(xml_entry)
            elif des_id.strip() in ['TRE_OVERFLOW', 'Registered Extensions', 'Controlled Extensions']:
                tres = TREList.from_bytes(data, 0)
                print_func('DESDATA = ')
                _print_tres(tres)
            else:
                # Unknown user-defined data
                print_func('DESDATA = {}'.format(_decode_effort(data)))
            print_func('')

    if details.res_subheader_offsets is not None:
        for res_subhead_num in range(details.res_subheader_offsets.size):
            print_func('----- RES {} -----'.format(res_subhead_num))
            hdr = details.parse_res_subheader(res_subhead_num)
            _print_res_header(hdr)
            data = details.get_res_bytes(res_subhead_num)
            print_func('RESDATA = {}'.format(_decode_effort(data)))
            print_func('')


##########
# method for dumping file using the print method(s)

def dump_nitf_file(file_name, dest, over_write=True):
    """
    Utility to dump the NITF header and various subheader details to a configurable
    destination.

    Parameters
    ----------
    file_name : str|BinaryIO
        The path to or file-like object containing a NITF 2.1 or 2.0 file.
    dest : str
        'stdout', 'string', 'default' (will use `file_name+'.header_dump.txt'`),
        or the path to an output file.
    over_write : bool
        If `True`, then overwrite the destination file, otherwise append to the
        file.

    Returns
    -------
    None|str
        There is only a return value if `dest=='string'`.
    """

    if dest == 'stdout':
        print_nitf(file_name, dest=sys.stdout)
        return
    if dest == 'string':
        out = StringIO()
        print_nitf(file_name, dest=out)
        value = out.getvalue()
        out.close()  # free the buffer
        return value

    the_out_file = _create_default_output_file(file_name) if dest == 'default' else dest
    if not os.path.exists(the_out_file) or over_write:
        with open(the_out_file, 'w') as the_file:
            print_nitf(file_name, dest=the_file)
    else:
        with open(the_out_file, 'a') as the_file:
            print_nitf(file_name, dest=the_file)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Utility to dump NITF 2.1 or 2.0 headers.',
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        'input_file',
        help='The path to a nitf file, or directory to search for NITF files.')
    parser.add_argument(
        '-o', '--output', default='default',
        help="'default', 'stdout', or the path for an output file.\n"
             "* 'default', the output will be at '<input path>.header_dump.txt' \n"
             "   This will be overwritten, if it exists.\n"
             "* 'stdout' will print the information to standard out.\n"
             "* Otherwise, "
             "     if `input_file` is a directory, this is expected to be the path to\n"
             "       an output directory for the output following the default naming scheme.\n"
             "*    if `input_file` a file path, this is expected to be the path to a file \n"
             "       and output will be written there.\n"
             "  In either case, existing output files will be overwritten.")
    args = parser.parse_args()

    if os.path.isdir(args.input_file):
        entries = [os.path.join(args.input_file, part) for part in os.listdir(args.input_file)]
        for file_number, entry in enumerate(filter(_filter_files, entries)):
            if args.output == 'stdout':
                output = args.output
            elif args.output == 'default':
                output = _create_default_output_file(entry, output_directory=None)
            else:
                if not os.path.isdir(args.output):
                    raise IOError(
                        'Provided input is a directory, so provided output must '
                        'be a directory, `stdout`, or `default`.')
                output = _create_default_output_file(entry, output_directory=args.output)
            dump_nitf_file(entry, output)
    else:
        dump_nitf_file(args.input_file, args.output)



"""
Definition for the FiducialInfo NGA modified RDE/AFRL labeling object
"""

__classification__ = "UNCLASSIFIED"
__authors__ = "Thomas McCullough"


from typing import Optional, List
import logging
import numpy

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import IntegerDescriptor, SerializableDescriptor, \
    SerializableListDescriptor, StringDescriptor
from sarpy.io.complex.sicd_elements.blocks import RowColType
from sarpy.io.complex.sicd_elements.SICD import SICDType

from .base import DEFAULT_STRICT
from .blocks import LatLonEleType, RangeCrossRangeType, \
    ProjectionPerturbationType, LabelSourceType

logger = logging.getLogger(__name__)

_no_projection_text = 'This sicd does not permit projection,\n\t' \
                      'so the image location can not be inferred'


class ImageLocationType(Serializable):
    _fields = ('CenterPixel', )
    _required = _fields
    # descriptors
    CenterPixel = SerializableDescriptor(
        'CenterPixel', RowColType, _required, strict=DEFAULT_STRICT,
        docstring='The pixel location of the center of the object')  # type: RowColType

    def __init__(self, CenterPixel=None, **kwargs):
        """
        Parameters
        ----------
        CenterPixel : RowColType|numpy.ndarray|list|tuple
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CenterPixel = CenterPixel
        super(ImageLocationType, self).__init__(**kwargs)

    @classmethod
    def from_geolocation(cls, geo_location, the_structure):
        """
        Construct from the corresponding Geolocation using the sicd
        projection model.

        Parameters
        ----------
        geo_location : GeoLocationType
        the_structure : SICDType|SIDDType

        Returns
        -------
        None|ImageLocationType
        """

        if geo_location is None or geo_location.CenterPixel is None:
            return None

        if not the_structure.can_project_coordinates():
            logger.warning(_no_projection_text)
            return None

        if isinstance(the_structure, SICDType):
            image_shift = numpy.array(
                [the_structure.ImageData.FirstRow, the_structure.ImageData.FirstCol], dtype='float64')
        else:
            image_shift = numpy.zeros((2, ), dtype='float64')

        absolute_pixel_location, _, _ = the_structure.project_ground_to_image_geo(
            geo_location.CenterPixel.get_array(dtype='float64'), ordering='latlong')
        if numpy.any(numpy.isnan(absolute_pixel_location)):
            return None

        return ImageLocationType(CenterPixel=absolute_pixel_location - image_shift)


class GeoLocationType(Serializable):
    _fields = ('CenterPixel', )
    _required = _fields
    # descriptors
    CenterPixel = SerializableDescriptor(
        'CenterPixel', LatLonEleType, _required, strict=DEFAULT_STRICT,
        docstring='The physical location of the center of the object')  # type: LatLonEleType

    def __init__(self, CenterPixel=None, **kwargs):
        """
        Parameters
        ----------
        CenterPixel : LatLonEleType|numpy.ndarray|list|tuple
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.CenterPixel = CenterPixel
        super(GeoLocationType, self).__init__(**kwargs)

    # noinspection PyUnusedLocal
    @classmethod
    def from_image_location(cls, image_location, the_structure, projection_type='HAE', **proj_kwargs):
        """
        Construct the geographical location from the image location via
        projection using the SICD model.

        .. Note::
            This assumes that the image coordinates are with respect to the given
            image (chip), and NOT including any sicd.ImageData.FirstRow/Col values,
            which will be added here.

        Parameters
        ----------
        image_location : ImageLocationType
        the_structure : SICDType|SIDDType
        projection_type : str
            The projection type selector, one of `['PLANE', 'HAE', 'DEM']`. Using `'DEM'`
            requires configuration for the DEM pathway described in
            :func:`sarpy.geometry.point_projection.image_to_ground_dem`.
        proj_kwargs
            The keyword arguments for the :func:`SICDType.project_image_to_ground_geo` method.

        Returns
        -------
        None|GeoLocationType
            Coordinates may be populated as `NaN` if projection fails.
        """

        if image_location is None or image_location.CenterPixel is None:
            return None

        if not the_structure.can_project_coordinates():
            logger.warning(_no_projection_text)
            return None

        # make sure this is defined, for the sake of efficiency
        the_structure.define_coa_projection(override=False)

        if isinstance(the_structure, SICDType):
            image_shift = numpy.array(
                [the_structure.ImageData.FirstRow, the_structure.ImageData.FirstCol], dtype='float64')
        else:
            image_shift = numpy.zeros((2, ), dtype='float64')

        coords = image_location.CenterPixel.get_array(dtype='float64') + image_shift
        geo_coords = the_structure.project_image_to_ground_geo(
            coords, ordering='latlong', projection_type=projection_type, **proj_kwargs)

        out = GeoLocationType(CenterPixel=geo_coords)
        return out


class PhysicalLocationType(Serializable):
    _fields = ('Physical', )
    _required = _fields
    # descriptors
    Physical = SerializableDescriptor(
        'Physical', ImageLocationType, _required, strict=DEFAULT_STRICT,
    )  # type: ImageLocationType

    def __init__(self, Physical=None, **kwargs):
        """
        Parameters
        ----------
        Physical : ImageLocationType
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Physical = Physical
        super(PhysicalLocationType, self).__init__(**kwargs)


class TheFiducialType(Serializable):
    _fields = (
        'Name', 'SerialNumber', 'FiducialType', 'DatasetFiducialNumber',
        'ImageLocation', 'GeoLocation',
        'IPRWidth3dB', 'IPRWidth18dB', 'IPRWidth3dB18dBRatio',
        'PeakSideLobeRatio', 'IntegratedSideLobeRatio',
        'SlantPlane', 'GroundPlane', 'ProjectionPerturbation')
    _required = (
        'FiducialType', 'ImageLocation', 'GeoLocation')
    # descriptors
    Name = StringDescriptor(
        'Name', _required, strict=DEFAULT_STRICT,
        docstring='Name of the fiducial.')  # type: Optional[str]
    SerialNumber = StringDescriptor(
        'SerialNumber', _required, strict=DEFAULT_STRICT,
        docstring='The serial number of the fiducial')  # type: Optional[str]
    FiducialType = StringDescriptor(
        'FiducialType',
        _required, strict=DEFAULT_STRICT,
        docstring='The type of fiducial')  # type: str
    DatasetFiducialNumber = IntegerDescriptor(
        'DatasetFiducialNumber', _required,
        docstring='Unique number of the fiducial within the selected dataset, '
                  'defined by the RDE system')  # type: Optional[int]
    ImageLocation = SerializableDescriptor(
        'ImageLocation', ImageLocationType, _required,
        docstring='Center of the fiducial in the image'
    )  # type: Optional[ImageLocationType]
    GeoLocation = SerializableDescriptor(
        'GeoLocation', GeoLocationType, _required,
        docstring='Real physical location of the fiducial'
    )  # type: Optional[GeoLocationType]
    IPRWidth3dB = SerializableDescriptor(
        'IPRWidth3dB', RangeCrossRangeType, _required,
        docstring='The 3 dB impulse response width, in meters'
    )  # type: Optional[RangeCrossRangeType]
    IPRWidth18dB = SerializableDescriptor(
        'IPRWidth18dB', RangeCrossRangeType, _required,
        docstring='The 18 dB impulse response width, in meters'
    )  # type: Optional[RangeCrossRangeType]
    IPRWidth3dB18dBRatio = SerializableDescriptor(
        'IPRWidth3dB18dBRatio', RangeCrossRangeType, _required,
        docstring='Ratio of the 3 dB to 18 dB system impulse response width'
    )  # type: Optional[RangeCrossRangeType]
    PeakSideLobeRatio = SerializableDescriptor(
        'PeakSideLobeRatio', RangeCrossRangeType, _required,
        docstring='Ratio of the peak sidelobe intensity to the peak mainlobe intensity, '
                  'in dB')  # type: Optional[RangeCrossRangeType]
    IntegratedSideLobeRatio = SerializableDescriptor(
        'IntegratedSideLobeRatio', RangeCrossRangeType, _required,
        docstring='Ratio of all the energies in the sidelobes of the '
                  'system impulse response to the energy in the mainlobe, '
                  'in dB')  # type: Optional[RangeCrossRangeType]
    SlantPlane = SerializableDescriptor(
        'SlantPlane', PhysicalLocationType, _required,
        docstring='Center of the object in the slant plane'
    )  # type: Optional[PhysicalLocationType]
    GroundPlane = SerializableDescriptor(
        'GroundPlane', PhysicalLocationType, _required,
        docstring='Center of the object in the ground plane'
    )  # type: Optional[PhysicalLocationType]
    ProjectionPerturbation = SerializableDescriptor(
        'ProjectionPerturbation', ProjectionPerturbationType, _required, 
        docstring='')  # type: Optional[ProjectionPerturbationType]

    def __init__(self, Name=None, SerialNumber=None, FiducialType=None,
                 DatasetFiducialNumber=None, ImageLocation=None, GeoLocation=None,
                 IPRWidth3dB=None, IPRWidth18dB=None, IPRWidth3dB18dBRatio=None,
                 PeakSideLobeRatio=None, IntegratedSideLobeRatio=None,
                 SlantPlane=None, GroundPlane=None, ProjectionPerturbation=None,
                 **kwargs):
        """
        Parameters
        ----------
        Name : str
        SerialNumber : None|str
        FiducialType : str
        DatasetFiducialNumber : None|int
        ImageLocation : ImageLocationType
        GeoLocation : GeoLocationType
        IPRWidth3dB : None|RangeCrossRangeType|numpy.ndarray|list|tuple
        IPRWidth18dB : None|RangeCrossRangeType|numpy.ndarray|list|tuple
        IPRWidth3dB18dBRatio : None|RangeCrossRangeType|numpy.ndarray|list|tuple
        PeakSideLobeRatio : None|RangeCrossRangeType|numpy.ndarray|list|tuple
        IntegratedSideLobeRatio : None|RangeCrossRangeType|numpy.ndarray|list|tuple
        SlantPlane : None|PhysicalLocationType
        GroundPlane : None|PhysicalLocationType
        ProjectionPerturbation : None|ProjectionPerturbationType
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Name = Name
        self.SerialNumber = SerialNumber
        self.FiducialType = FiducialType
        self.DatasetFiducialNumber = DatasetFiducialNumber
        self.ImageLocation = ImageLocation
        self.GeoLocation = GeoLocation
        self.IPRWidth3dB = IPRWidth3dB
        self.IPRWidth18dB = IPRWidth18dB
        self.IPRWidth3dB18dBRatio = IPRWidth3dB18dBRatio
        self.PeakSideLobeRatio = PeakSideLobeRatio
        self.IntegratedSideLobeRatio = IntegratedSideLobeRatio
        self.SlantPlane = SlantPlane
        self.GroundPlane = GroundPlane
        self.ProjectionPerturbation = ProjectionPerturbation
        super(TheFiducialType, self).__init__(**kwargs)

    def set_image_location_from_sicd(self, sicd, populate_in_periphery=False):
        """
        Set the image location information with respect to the given SICD.

        Parameters
        ----------
        sicd : SICDType
        populate_in_periphery : bool

        Returns
        -------
        int
            -1 - insufficient metadata to proceed
            0 - nothing to be done
            1 - successful
            2 - object in image periphery, not populating
            3 - object not in image field
        """

        if self.ImageLocation is not None or self.SlantPlane is not None:
            # no need to infer anything, it's already populated
            return 0

        if self.GeoLocation is None:
            logger.warning(
                'GeoLocation is not populated,\n\t'
                'so the image location can not be inferred')
            return -1

        if not sicd.can_project_coordinates():
            logger.warning(_no_projection_text)
            return -1

        image_location = ImageLocationType.from_geolocation(self.GeoLocation, sicd)
        # check bounding information
        rows = sicd.ImageData.NumRows
        cols = sicd.ImageData.NumCols
        center_pixel = image_location.CenterPixel.get_array(dtype='float64')

        if (0 < center_pixel[0] < rows - 1) and (0 < center_pixel[1] < cols - 1):
            placement = 1
        elif (-3 < center_pixel[0] < rows + 2) and (-3 < center_pixel[1] < cols + 2):
            placement = 2
        else:
            placement = 3

        if placement == 3 or (placement == 2 and not populate_in_periphery):
            return placement

        self.ImageLocation = image_location
        self.SlantPlane = PhysicalLocationType(Physical=image_location)
        return placement

    def set_geo_location_from_sicd(self, sicd, projection_type='HAE', **proj_kwargs):
        """
        Set the geographical location information with respect to the given SICD,
        assuming that the image coordinates are populated.

        .. Note::
            This assumes that the image coordinates are with respect to the given
            image (chip), and NOT including any sicd.ImageData.FirstRow/Col values,
            which will be added here.

        Parameters
        ----------
        sicd : SICDType
        projection_type : str
            The projection type selector, one of `['PLANE', 'HAE', 'DEM']`. Using `'DEM'`
            requires configuration for the DEM pathway described in
            :func:`sarpy.geometry.point_projection.image_to_ground_dem`.
        proj_kwargs
            The keyword arguments for the :func:`SICDType.project_image_to_ground_geo` method.
        """

        if self.GeoLocation is not None:
            # no need to infer anything, it's already populated
            return

        if self.ImageLocation is None:
            logger.warning(
                'ImageLocation is not populated,\n\t'
                'so the geographical location can not be inferred')
            return

        if not sicd.can_project_coordinates():
            logger.warning(_no_projection_text)
            return

        self.GeoLocation = GeoLocationType.from_image_location(
            self.ImageLocation, sicd, projection_type=projection_type, **proj_kwargs)


class FiducialInfoType(Serializable):
    _fields = (
        'NumberOfFiducialsInImage', 'NumberOfFiducialsInScene', 'LabelSource', 'Fiducials')
    _required = (
        'NumberOfFiducialsInImage', 'NumberOfFiducialsInScene', 'LabelSource', 'Fiducials')
    _collections_tags = {'Fiducials': {'array': False, 'child_tag': 'Fiducial'}}
    # descriptors
    NumberOfFiducialsInImage = IntegerDescriptor(
        'NumberOfFiducialsInImage', _required, strict=DEFAULT_STRICT,
        docstring='Number of ground truthed objects in the image.')  # type: int
    NumberOfFiducialsInScene = IntegerDescriptor(
        'NumberOfFiducialsInScene', _required, strict=DEFAULT_STRICT,
        docstring='Number of ground truthed objects in the scene.')  # type: int
    LabelSource = SerializableDescriptor(
        'LabelSource', LabelSourceType, _required, strict=DEFAULT_STRICT,
        docstring='The source of the labels')  # type: LabelSourceType
    Fiducials = SerializableListDescriptor(
        'Fiducials', TheFiducialType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='The object collection')  # type: List[TheFiducialType]

    def __init__(self, NumberOfFiducialsInImage=None, NumberOfFiducialsInScene=None,
                 LabelSource=None, Fiducials=None, **kwargs):
        """
        Parameters
        ----------
        NumberOfFiducialsInImage : int
        NumberOfFiducialsInScene : int
        LabelSource : LabelSourceType
        Fiducials : None|List[TheFiducialType]
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.NumberOfFiducialsInImage = NumberOfFiducialsInImage
        self.NumberOfFiducialsInScene = NumberOfFiducialsInScene
        self.LabelSource = LabelSource
        self.Fiducials = Fiducials
        super(FiducialInfoType, self).__init__(**kwargs)

    def set_image_location_from_sicd(
            self, sicd, populate_in_periphery=False, include_out_of_range=False):
        """
        Set the image location information with respect to the given SICD,
        assuming that the physical coordinates are populated. The `NumberOfFiducialsInImage`
        will be set, and `NumberOfFiducialsInScene` will be left unchanged.

        Parameters
        ----------
        sicd : SICDType
        populate_in_periphery : bool
            Populate image information for objects on the periphery?
        include_out_of_range : bool
            Include the objects which are out of range (with no image location information)?
        """

        def update_fiducial(temp_fid, in_image_count):
            status = temp_fid.set_image_location_from_sicd(sicd, populate_in_periphery=populate_in_periphery)
            use_fid = False
            if status == 0:
                raise ValueError('Fiducial already has image details set')
            if status == 1 or (status == 2 and populate_in_periphery):
                use_fid = True
                in_image_count += 1
            return use_fid, in_image_count

        fid_in_image = 0
        if include_out_of_range:
            # the fiducials list is just modified in place
            for the_fid in self.Fiducials:
                _, fid_in_image = update_fiducial(the_fid, fid_in_image)
        else:
            # the fiducials list is just modified in place
            fiducials = []
            for the_fid in self.Fiducials:
                use_this_fid, fid_in_image = update_fiducial(the_fid, fid_in_image)
                if use_this_fid:
                    fiducials.append(the_fid)
            self.Fiducials = fiducials
        self.NumberOfFiducialsInImage = fid_in_image



"""
Work in progress for reading some other kind of complex NITF.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
from typing import Union, Tuple, List, Optional, Callable, Sequence
import copy
from datetime import datetime

import numpy
from scipy.constants import foot

from sarpy.geometry.geocoords import geodetic_to_ecf, ned_to_ecf
from sarpy.geometry.latlon import num as lat_lon_parser

from sarpy.io.general.base import SarpyIOError
from sarpy.io.general.data_segment import DataSegment, SubsetSegment
from sarpy.io.general.format_function import FormatFunction, ComplexFormatFunction
from sarpy.io.general.nitf import extract_image_corners, NITFDetails, NITFReader
from sarpy.io.general.nitf_elements.security import NITFSecurityTags
from sarpy.io.general.nitf_elements.image import ImageSegmentHeader, ImageSegmentHeader0
from sarpy.io.general.nitf_elements.nitf_head import NITFHeader, NITFHeader0
from sarpy.io.general.nitf_elements.base import TREList
from sarpy.io.general.nitf_elements.tres.unclass.CMETAA import CMETAA
from sarpy.io.general.utils import is_file_like

from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.CollectionInfo import CollectionInfoType
from sarpy.io.complex.sicd_elements.ImageData import ImageDataType
from sarpy.io.complex.sicd_elements.GeoData import GeoDataType, SCPType
from sarpy.io.complex.sicd_elements.Grid import GridType, DirParamType, WgtTypeType
from sarpy.io.complex.sicd_elements.Timeline import TimelineType, IPPSetType
from sarpy.io.complex.sicd_elements.RadarCollection import RadarCollectionType, \
    TxFrequencyType, WaveformParametersType, ChanParametersType
from sarpy.io.complex.sicd_elements.SCPCOA import SCPCOAType
from sarpy.io.complex.sicd_elements.ImageFormation import ImageFormationType, TxFrequencyProcType
from sarpy.io.complex.sicd_elements.ImageCreation import ImageCreationType
from sarpy.io.complex.sicd_elements.PFA import PFAType


logger = logging.getLogger(__name__)

_iso_date_format = '{}-{}-{}T{}:{}:{}'

# NB: DO NOT implement is_a() here.
#   This will explicitly happen after other readers


########
# Define sicd structure from image sub-header information

def extract_sicd(
        img_header: Union[ImageSegmentHeader, ImageSegmentHeader0],
        transpose: True,
        nitf_header: Optional[Union[NITFHeader, NITFHeader0]] = None) -> SICDType:
    """
    Extract the best available SICD structure from relevant nitf header structures.

    Parameters
    ----------
    img_header : ImageSegmentHeader|ImageSegmentHeader0
    transpose : bool
    nitf_header : None|NITFHeader|NITFHeader0

    Returns
    -------
    SICDType
    """

    def get_collection_info() -> CollectionInfoType:
        isorce = img_header.ISORCE.strip()
        collector_name = None if len(isorce) < 1 else isorce

        iid2 = img_header.IID2.strip()
        core_name = img_header.IID1.strip() if len(iid2) < 1 else iid2

        class_str = img_header.Security.CLAS
        if class_str == 'T':
            classification = 'TOPSECRET'
        elif class_str == 'S':
            classification = 'SECRET'
        elif class_str == 'C':
            classification = 'CONFIDENTIAL'
        elif class_str == 'U':
            classification = 'UNCLASSIFIED'
        else:
            classification = ''
        ctlh = img_header.Security.CTLH.strip()
        if len(ctlh) < 1:
            classification += '//' + ctlh
        code = img_header.Security.CODE.strip()
        if len(code) < 1:
            classification += '//' + code

        return CollectionInfoType(
            CollectorName=collector_name,
            CoreName=core_name,
            Classification=classification)

    def get_image_data() -> ImageDataType:
        pvtype = img_header.PVTYPE
        if pvtype == 'C':
            if img_header.NBPP != 64:
                logger.warning(
                    'This NITF has complex bands that are not 64-bit.\n\t'
                    'This is not currently supported.')
            pixel_type = 'RE32F_IM32F'
        elif pvtype == 'R':
            if img_header.NBPP == 64:
                logger.warning(
                    'The real/imaginary data in the NITF are stored as 64-bit floating point.\n\t'
                    'The closest Pixel Type, RE32F_IM32F, will be used,\n\t'
                    'but there may be overflow issues if converting this file.')
            pixel_type = 'RE32F_IM32F'
        elif pvtype == 'SI':
            pixel_type = 'RE16I_IM16I'
        else:
            raise ValueError('Got unhandled PVTYPE {}'.format(pvtype))

        if transpose:
            rows = img_header.NCOLS
            cols = img_header.NROWS
        else:
            rows = img_header.NROWS
            cols = img_header.NCOLS
        return ImageDataType(
            PixelType=pixel_type,
            NumRows=rows,
            NumCols=cols,
            FirstRow=0,
            FirstCol=0,
            FullImage=(rows, cols),
            SCPPixel=(0.5 * rows, 0.5 * cols))

    def append_country_code(cc) -> None:
        if len(cc) > 0:
            if the_sicd.CollectionInfo is None:
                the_sicd.CollectionInfo = CollectionInfoType(CountryCodes=[cc, ])
            elif the_sicd.CollectionInfo.CountryCodes is None:
                the_sicd.CollectionInfo.CountryCodes = [cc, ]
            elif cc not in the_sicd.CollectionInfo.CountryCodes:
                the_sicd.CollectionInfo.CountryCodes.append(cc)

    def set_image_corners(icps: numpy.ndarray, override: bool = False) -> None:
        if the_sicd.GeoData is None:
            the_sicd.GeoData = GeoDataType(ImageCorners=icps)
        elif the_sicd.GeoData.ImageCorners is None or override:
            the_sicd.GeoData.ImageCorners = icps

    def set_arp_position(arp_ecf: numpy.ndarray, override: bool = False) -> None:
        if the_sicd.SCPCOA is None:
            the_sicd.SCPCOA = SCPCOAType(ARPPos=arp_ecf)
        elif override:
            # prioritize this information first - it should be more reliable than other sources
            the_sicd.SCPCOA.ARPPos = arp_ecf

    def set_scp(scp_ecf: numpy.ndarray, scp_pixel: Union[numpy.ndarray, list, tuple], override: bool = False) -> None:
        def set_scppixel():
            if the_sicd.ImageData is None:
                the_sicd.ImageData = ImageDataType(SCPPixel=scp_pixel)
            else:
                the_sicd.ImageData.SCPPixel = scp_pixel
        if the_sicd.GeoData is None:
            the_sicd.GeoData = GeoDataType(SCP=SCPType(ECF=scp_ecf))
            set_scppixel()
        elif the_sicd.GeoData.SCP is None or override:
            the_sicd.GeoData.SCP = SCPType(ECF=scp_ecf)
            set_scppixel()

    def set_collect_start(
            collect_start: Union[str, datetime, numpy.datetime64], override: bool = False) -> None:
        if the_sicd.Timeline is None:
            the_sicd.Timeline = TimelineType(CollectStart=collect_start)
        elif the_sicd.Timeline.CollectStart is None or override:
            the_sicd.Timeline.CollectStart = collect_start

    def set_uvects(row_unit: numpy.ndarray, col_unit: numpy.ndarray) -> None:
        if the_sicd.Grid is None:
            the_sicd.Grid = GridType(
                Row=DirParamType(UVectECF=row_unit),
                Col=DirParamType(UVectECF=col_unit))
            return

        if the_sicd.Grid.Row is None:
            the_sicd.Grid.Row = DirParamType(UVectECF=row_unit)
        elif the_sicd.Grid.Row.UVectECF is None:
            the_sicd.Grid.Row.UVectECF = row_unit

        if the_sicd.Grid.Col is None:
            the_sicd.Grid.Col = DirParamType(UVectECF=col_unit)
        elif the_sicd.Grid.Col.UVectECF is None:
            the_sicd.Grid.Col.UVectECF = col_unit

    def try_CMETAA() -> None:
        # noinspection PyTypeChecker
        tre = None if tres is None else tres['CMETAA']  # type: CMETAA
        if tre is None:
            return

        cmetaa = tre.DATA

        if the_sicd.GeoData is None:
            the_sicd.GeoData = GeoDataType()
        if the_sicd.SCPCOA is None:
            the_sicd.SCPCOA = SCPCOAType()
        if the_sicd.Grid is None:
            the_sicd.Grid = GridType()
        if the_sicd.Timeline is None:
            the_sicd.Timeline = TimelineType()
        if the_sicd.RadarCollection is None:
            the_sicd.RadarCollection = RadarCollectionType()
        if the_sicd.ImageFormation is None:
            the_sicd.ImageFormation = ImageFormationType()

        the_sicd.SCPCOA.SCPTime = 0.5*float(cmetaa.WF_CDP)
        the_sicd.GeoData.SCP = SCPType(ECF=tre.get_scp())
        the_sicd.SCPCOA.ARPPos = tre.get_arp()

        the_sicd.SCPCOA.SideOfTrack = cmetaa.CG_LD.strip().upper()
        the_sicd.SCPCOA.SlantRange = float(cmetaa.CG_SRAC)
        the_sicd.SCPCOA.DopplerConeAng = float(cmetaa.CG_CAAC)
        the_sicd.SCPCOA.GrazeAng = float(cmetaa.CG_GAAC)
        the_sicd.SCPCOA.IncidenceAng = 90 - float(cmetaa.CG_GAAC)
        if hasattr(cmetaa, 'CG_TILT'):
            the_sicd.SCPCOA.TwistAng = float(cmetaa.CG_TILT)
        if hasattr(cmetaa, 'CG_SLOPE'):
            the_sicd.SCPCOA.SlopeAng = float(cmetaa.CG_SLOPE)

        the_sicd.ImageData.SCPPixel = [int(cmetaa.IF_DC_IS_COL), int(cmetaa.IF_DC_IS_ROW)]
        img_corners = tre.get_image_corners()
        if img_corners is not None:
            the_sicd.GeoData.ImageCorners = img_corners

        if cmetaa.CMPLX_SIGNAL_PLANE.upper() == 'S':
            the_sicd.Grid.ImagePlane = 'SLANT'
        elif cmetaa.CMPLX_SIGNAL_PLANE.upper() == 'G':
            the_sicd.Grid.ImagePlane = 'GROUND'
        else:
            logger.warning(
                'Got unexpected CMPLX_SIGNAL_PLANE value {},\n\t'
                'setting ImagePlane to SLANT'.format(cmetaa.CMPLX_SIGNAL_PLANE))

        the_sicd.Grid.Row = DirParamType(
            SS=float(cmetaa.IF_RSS),
            ImpRespWid=float(cmetaa.IF_RGRES),
            Sgn=1 if cmetaa.IF_RFFTS.strip() == '-' else -1,  # opposite sign convention
            ImpRespBW=float(cmetaa.IF_RFFT_SAMP)/(float(cmetaa.IF_RSS)*float(cmetaa.IF_RFFT_TOT)))
        the_sicd.Grid.Col = DirParamType(
            SS=float(cmetaa.IF_AZSS),
            ImpRespWid=float(cmetaa.IF_AZRES),
            Sgn=1 if cmetaa.IF_AFFTS.strip() == '-' else -1,  # opposite sign convention
            ImpRespBW=float(cmetaa.IF_AZFFT_SAMP)/(float(cmetaa.IF_AZSS)*float(cmetaa.IF_AZFFT_TOT)))
        cmplx_weight = cmetaa.CMPLX_WEIGHT.strip().upper()
        if cmplx_weight == 'UWT':
            the_sicd.Grid.Row.WgtType = WgtTypeType(WindowName='UNIFORM')
            the_sicd.Grid.Col.WgtType = WgtTypeType(WindowName='UNIFORM')
        elif cmplx_weight == 'HMW':
            the_sicd.Grid.Row.WgtType = WgtTypeType(WindowName='HAMMING')
            the_sicd.Grid.Col.WgtType = WgtTypeType(WindowName='HAMMING')
        elif cmplx_weight == 'HNW':
            the_sicd.Grid.Row.WgtType = WgtTypeType(WindowName='HANNING')
            the_sicd.Grid.Col.WgtType = WgtTypeType(WindowName='HANNING')
        elif cmplx_weight == 'TAY':
            the_sicd.Grid.Row.WgtType = WgtTypeType(
                WindowName='TAYLOR',
                Parameters={
                    'SLL': '-{0:d}'.format(int(cmetaa.CMPLX_RNG_SLL)),
                    'NBAR': '{0:d}'.format(int(cmetaa.CMPLX_RNG_TAY_NBAR))})
            the_sicd.Grid.Col.WgtType = WgtTypeType(
                WindowName='TAYLOR',
                Parameters={
                    'SLL': '-{0:d}'.format(int(cmetaa.CMPLX_AZ_SLL)),
                    'NBAR': '{0:d}'.format(int(cmetaa.CMPLX_AZ_TAY_NBAR))})
        else:
            logger.warning(
                'Got unsupported CMPLX_WEIGHT value {}.\n\tThe resulting SICD will '
                'not have valid weight array populated'.format(cmplx_weight))
        the_sicd.Grid.Row.define_weight_function()
        the_sicd.Grid.Col.define_weight_function()

        # noinspection PyBroadException
        try:
            date_str = cmetaa.T_UTC_YYYYMMMDD
            time_str = cmetaa.T_HHMMSSUTC
            date_time = _iso_date_format.format(
                date_str[:4], date_str[4:6], date_str[6:8],
                time_str[:2], time_str[2:4], time_str[4:6])
            the_sicd.Timeline.CollectStart = numpy.datetime64(date_time, 'us')
        except Exception:
            logger.info('Failed extracting start time from CMETAA')
            pass
        the_sicd.Timeline.CollectDuration = float(cmetaa.WF_CDP)
        the_sicd.Timeline.IPP = [
            IPPSetType(TStart=0,
                       TEnd=float(cmetaa.WF_CDP),
                       IPPStart=0,
                       IPPEnd=numpy.floor(float(cmetaa.WF_CDP)*float(cmetaa.WF_PRF)),
                       IPPPoly=[0, float(cmetaa.WF_PRF)])]

        the_sicd.RadarCollection.TxFrequency = TxFrequencyType(
            Min=float(cmetaa.WF_SRTFR),
            Max=float(cmetaa.WF_ENDFR))
        the_sicd.RadarCollection.TxPolarization = cmetaa.POL_TR.upper()
        the_sicd.RadarCollection.Waveform = [WaveformParametersType(
            TxPulseLength=float(cmetaa.WF_WIDTH),
            TxRFBandwidth=float(cmetaa.WF_BW),
            TxFreqStart=float(cmetaa.WF_SRTFR),
            TxFMRate=float(cmetaa.WF_CHRPRT)*1e12)]
        tx_rcv_pol = '{}:{}'.format(cmetaa.POL_TR.upper(), cmetaa.POL_RE.upper())
        the_sicd.RadarCollection.RcvChannels = [
            ChanParametersType(TxRcvPolarization=tx_rcv_pol)]

        the_sicd.ImageFormation.TxRcvPolarizationProc = tx_rcv_pol
        if_process = cmetaa.IF_PROCESS.strip().upper()
        if if_process == 'PF':
            the_sicd.ImageFormation.ImageFormAlgo = 'PFA'
            scp_ecf = tre.get_scp()
            fpn_ned = numpy.array(
                [float(cmetaa.CG_FPNUV_X), float(cmetaa.CG_FPNUV_Y), float(cmetaa.CG_FPNUV_Z)], dtype='float64')
            ipn_ned = numpy.array(
                [float(cmetaa.CG_IDPNUVX), float(cmetaa.CG_IDPNUVY), float(cmetaa.CG_IDPNUVZ)], dtype='float64')
            fpn_ecf = ned_to_ecf(fpn_ned, scp_ecf, absolute_coords=False)
            ipn_ecf = ned_to_ecf(ipn_ned, scp_ecf, absolute_coords=False)
            the_sicd.PFA = PFAType(FPN=fpn_ecf, IPN=ipn_ecf)
        elif if_process in ['RM', 'CD']:
            the_sicd.ImageFormation.ImageFormAlgo = 'RMA'

        # the remainder of this is guesswork to define required fields
        the_sicd.ImageFormation.TStartProc = 0  # guess work
        the_sicd.ImageFormation.TEndProc = float(cmetaa.WF_CDP)
        the_sicd.ImageFormation.TxFrequencyProc = TxFrequencyProcType(
            MinProc=float(cmetaa.WF_SRTFR), MaxProc=float(cmetaa.WF_ENDFR))
        # all remaining guess work
        the_sicd.ImageFormation.STBeamComp = 'NO'
        the_sicd.ImageFormation.ImageBeamComp = 'SV' if cmetaa.IF_BEAM_COMP[0] == 'Y' else 'NO'
        the_sicd.ImageFormation.AzAutofocus = 'NO' if cmetaa.AF_TYPE[0] == 'N' else 'SV'
        the_sicd.ImageFormation.RgAutofocus = 'NO'

    def try_AIMIDA() -> None:
        tre = None if tres is None else tres['AIMIDA']
        if tre is None:
            return
        aimida = tre.DATA

        append_country_code(aimida.COUNTRY.strip())

        create_time = datetime.strptime(aimida.CREATION_DATE, '%d%b%y')
        if the_sicd.ImageCreation is None:
            the_sicd.ImageCreation = ImageCreationType(DateTime=create_time)
        elif the_sicd.ImageCreation.DateTime is None:
            the_sicd.ImageCreation.DateTime = create_time

        collect_start = datetime.strptime(aimida.MISSION_DATE+aimida.TIME, '%d%b%y%H%M')
        set_collect_start(collect_start, override=False)

    def try_AIMIDB() -> None:
        tre = None if tres is None else tres['AIMIDB']
        if tre is None:
            return
        aimidb = tre.DATA

        append_country_code(aimidb.COUNTRY.strip())

        if the_sicd.ImageFormation is not None and the_sicd.ImageFormation.SegmentIdentifier is None:
            the_sicd.ImageFormation.SegmentIdentifier = aimidb.CURRENT_SEGMENT.strip()

        date_str = aimidb.ACQUISITION_DATE
        collect_start = numpy.datetime64(_iso_date_format.format(
            date_str[:4], date_str[4:6], date_str[6:8],
            date_str[8:10], date_str[10:12], date_str[12:14]), 'us')
        set_collect_start(collect_start, override=False)

    def try_ACFT() -> None:
        if tres is None:
            return
        tre = tres['ACFTA']
        if tre is None:
            tre = tres['ACFTB']
        if tre is None:
            return
        acft = tre.DATA

        sensor_id = acft.SENSOR_ID.strip()
        if len(sensor_id) > 1:
            if the_sicd.CollectionInfo is None:
                the_sicd.CollectionInfo = CollectionInfoType(CollectorName=sensor_id)
            elif the_sicd.CollectionInfo.CollectorName is None:
                the_sicd.CollectionInfo.CollectorName = sensor_id

        row_ss = float(acft.ROW_SPACING)
        col_ss = float(acft.COL_SPACING)

        if hasattr(acft, 'ROW_SPACING_UNITS') and acft.ROW_SPACING_UNITS.strip().lower() == 'f':
            row_ss *= foot
        if hasattr(acft, 'COL_SPACING_UNITS') and acft.COL_SPACING_UNITS.strip().lower() == 'f':
            col_ss *= foot

        # NB: these values are actually ground plane values, and should be
        # corrected to slant plane if possible
        if the_sicd.SCPCOA is not None:
            if the_sicd.SCPCOA.GrazeAng is not None:
                col_ss *= numpy.cos(numpy.deg2rad(the_sicd.SCPCOA.GrazeAng))
            if the_sicd.SCPCOA.TwistAng is not None:
                row_ss *= numpy.cos(numpy.deg2rad(the_sicd.SCPCOA.TwistAng))

        if the_sicd.Grid is None:
            the_sicd.Grid = GridType(Row=DirParamType(SS=row_ss), Col=DirParamType(SS=col_ss))
            return

        if the_sicd.Grid.Row is None:
            the_sicd.Grid.Row = DirParamType(SS=row_ss)
        elif the_sicd.Grid.Row.SS is None:
            the_sicd.Grid.Row.SS = row_ss

        if the_sicd.Grid.Col is None:
            the_sicd.Grid.Col = DirParamType(SS=col_ss)
        elif the_sicd.Grid.Col.SS is None:
            the_sicd.Grid.Col.SS = col_ss

    def try_BLOCKA() -> None:
        tre = None if tres is None else tres['BLOCKA']
        if tre is None:
            return
        blocka = tre.DATA

        icps = []
        for fld_name in ['FRFC_LOC', 'FRLC_LOC', 'LRLC_LOC', 'LRFC_LOC']:
            value = getattr(blocka, fld_name)
            # noinspection PyBroadException
            try:
                lat_val = float(value[:10])
                lon_val = float(value[10:21])
            except ValueError:
                lat_val = lat_lon_parser(value[:10])
                lon_val = lat_lon_parser(value[10:21])

            icps.append([lat_val, lon_val])
        set_image_corners(numpy.array(icps, dtype='float64'), override=False)

    def try_MPDSRA() -> None:
        def valid_array(arr):
            return numpy.all(numpy.isfinite(arr)) and numpy.any(arr != 0)

        tre = None if tres is None else tres['MPDSRA']
        if tre is None:
            return
        mpdsra = tre.DATA

        scp_ecf = foot*numpy.array(
            [float(mpdsra.ORO_X), float(mpdsra.ORO_Y), float(mpdsra.ORO_Z)], dtype='float64')
        if valid_array(scp_ecf):
            set_scp(scp_ecf, (int(mpdsra.ORP_COLUMN) - 1, int(mpdsra.ORP_ROW) - 1), override=False)

        arp_pos_ned = foot*numpy.array(
            [float(mpdsra.ARP_POS_N), float(mpdsra.ARP_POS_E), float(mpdsra.ARP_POS_D)], dtype='float64')
        arp_vel_ned = foot*numpy.array(
            [float(mpdsra.ARP_VEL_N), float(mpdsra.ARP_VEL_E), float(mpdsra.ARP_VEL_D)], dtype='float64')
        arp_acc_ned = foot*numpy.array(
            [float(mpdsra.ARP_ACC_N), float(mpdsra.ARP_ACC_E), float(mpdsra.ARP_ACC_D)], dtype='float64')
        arp_pos = ned_to_ecf(arp_pos_ned, scp_ecf, absolute_coords=True) if valid_array(arp_pos_ned) else None
        set_arp_position(arp_pos, override=False)

        arp_vel = ned_to_ecf(arp_vel_ned, scp_ecf, absolute_coords=False) if valid_array(arp_vel_ned) else None
        if the_sicd.SCPCOA.ARPVel is None:
            the_sicd.SCPCOA.ARPVel = arp_vel
        arp_acc = ned_to_ecf(arp_acc_ned, scp_ecf, absolute_coords=False) if valid_array(arp_acc_ned) else None
        if the_sicd.SCPCOA.ARPAcc is None:
            the_sicd.SCPCOA.ARPAcc = arp_acc

        if the_sicd.PFA is not None and the_sicd.PFA.FPN is None:
            # TODO: is this already in meters?
            fpn_ecf = numpy.array(
                [float(mpdsra.FOC_X), float(mpdsra.FOC_Y), float(mpdsra.FOC_Z)], dtype='float64')  # *foot
            if valid_array(fpn_ecf):
                the_sicd.PFA.FPN = fpn_ecf

    def try_MENSRB() -> None:
        tre = None if tres is None else tres['MENSRB']
        if tre is None:
            return
        mensrb = tre.DATA

        arp_llh = numpy.array(
            [lat_lon_parser(mensrb.ACFT_LOC[:12]),
             lat_lon_parser(mensrb.ACFT_LOC[12:25]),
             foot*float(mensrb.ACFT_ALT)], dtype='float64')
        scp_llh = numpy.array(
            [lat_lon_parser(mensrb.RP_LOC[:12]),
             lat_lon_parser(mensrb.RP_LOC[12:25]),
             foot*float(mensrb.RP_ELV)], dtype='float64')
        # TODO: handle the conversion from msl to hae

        arp_ecf = geodetic_to_ecf(arp_llh)
        scp_ecf = geodetic_to_ecf(scp_llh)
        set_arp_position(arp_ecf, override=True)

        set_scp(scp_ecf, (int(mensrb.RP_COL)-1, int(mensrb.RP_ROW)-1), override=False)

        row_unit_ned = numpy.array(
            [float(mensrb.C_R_NC), float(mensrb.C_R_EC), float(mensrb.C_R_DC)], dtype='float64')
        col_unit_ned = numpy.array(
            [float(mensrb.C_AZ_NC), float(mensrb.C_AZ_EC), float(mensrb.C_AZ_DC)], dtype='float64')
        set_uvects(ned_to_ecf(row_unit_ned, scp_ecf, absolute_coords=False),
                   ned_to_ecf(col_unit_ned, scp_ecf, absolute_coords=False))

    def try_MENSRA() -> None:
        tre = None if tres is None else tres['MENSRA']
        if tre is None:
            return
        mensra = tre.DATA

        arp_llh = numpy.array(
            [lat_lon_parser(mensra.ACFT_LOC[:10]),
             lat_lon_parser(mensra.ACFT_LOC[10:21]),
             foot*float(mensra.ACFT_ALT)], dtype='float64')
        scp_llh = numpy.array(
            [lat_lon_parser(mensra.CP_LOC[:10]),
             lat_lon_parser(mensra.CP_LOC[10:21]),
             foot*float(mensra.CP_ALT)], dtype='float64')
        # TODO: handle the conversion from msl to hae

        arp_ecf = geodetic_to_ecf(arp_llh)
        scp_ecf = geodetic_to_ecf(scp_llh)
        set_arp_position(arp_ecf, override=True)

        # TODO: is this already zero based?
        set_scp(geodetic_to_ecf(scp_llh), (int(mensra.CCRP_COL), int(mensra.CCRP_ROW)), override=False)

        row_unit_ned = numpy.array(
            [float(mensra.C_R_NC), float(mensra.C_R_EC), float(mensra.C_R_DC)], dtype='float64')
        col_unit_ned = numpy.array(
            [float(mensra.C_AZ_NC), float(mensra.C_AZ_EC), float(mensra.C_AZ_DC)], dtype='float64')
        set_uvects(ned_to_ecf(row_unit_ned, scp_ecf, absolute_coords=False),
                   ned_to_ecf(col_unit_ned, scp_ecf, absolute_coords=False))

    def extract_corners() -> None:
        icps = extract_image_corners(img_header)
        if icps is None:
            return
        # TODO: include symmetry transform issue
        set_image_corners(icps, override=False)

    def extract_start() -> None:
        # noinspection PyBroadException
        try:
            date_str = img_header.IDATIM
            collect_start = numpy.datetime64(
                _iso_date_format.format(
                    date_str[:4], date_str[4:6], date_str[6:8],
                    date_str[8:10], date_str[10:12], date_str[12:14]), 'us')
        except Exception:
            logger.info('failed extracting start time from IDATIM tre')
            return

        set_collect_start(collect_start, override=False)

    # noinspection PyUnresolvedReferences
    tres = None if img_header.ExtendedHeader.data is None \
        else img_header.ExtendedHeader.data  # type: Union[None, TREList]

    collection_info = get_collection_info()
    image_data = get_image_data()
    the_sicd = SICDType(
        CollectionInfo=collection_info,
        ImageData=image_data)
    # apply the various tres and associated logic
    # NB: this should generally be in order of preference
    try_CMETAA()
    try_AIMIDB()
    try_AIMIDA()
    try_ACFT()
    try_BLOCKA()
    try_MPDSRA()
    try_MENSRA()
    try_MENSRB()
    extract_corners()
    extract_start()
    return the_sicd


# Helper methods for transforming data

def get_linear_magnitude_scaling(scale_factor: float):
    """
    Get a linear magnitude scaling function, to correct magnitude.

    Parameters
    ----------
    scale_factor : float
        The scale factor, according to the definition given in STDI-0002.

    Returns
    -------
    callable
    """

    def scaler(data):
        return data/scale_factor
    return scaler


def get_linear_power_scaling(scale_factor):
    """
    Get a linear power scaling function, to derive correct magnitude.

    Parameters
    ----------
    scale_factor : float
        The scale factor, according to the definition given in STDI-0002.

    Returns
    -------
    callable
    """

    def scaler(data):
        return numpy.sqrt(data/scale_factor)
    return scaler


def get_log_magnitude_scaling(scale_factor, db_per_step):
    """
    Gets the log magnitude scaling function, to derive correct magnitude.

    Parameters
    ----------
    scale_factor : float
        The scale factor, according to the definition given in STDI-0002.
    db_per_step : float
        The db_per_step factor, according to the definiton given in STDI-0002

    Returns
    -------
    callable
    """

    lin_scaler = get_linear_magnitude_scaling(scale_factor)

    def scaler(data):
        return lin_scaler(numpy.exp(0.05*numpy.log(10)*db_per_step*data))

    return scaler


def get_log_power_scaling(scale_factor, db_per_step):
    """
    Gets the log power scaling function, to derive correct magnitude.

    Parameters
    ----------
    scale_factor : float
        The scale factor, according to the definition given in STDI-0002.
    db_per_step : float
        The db_per_step factor, according to the definiton given in STDI-0002

    Returns
    -------
    callable
    """

    power_scaler = get_linear_power_scaling(scale_factor)

    def scaler(data):
        return power_scaler(numpy.exp(0.1*numpy.log(10)*db_per_step*data))

    return scaler


def get_linlog_magnitude_scaling(scale_factor, tipping_point):
    """
    Gets the magnitude scaling function for the model which
    is initially linear, and then switches to logarithmic beyond a fixed
    tipping point.

    Parameters
    ----------
    scale_factor : float
        The scale factor, according to the definition given in STDI-0002.
    tipping_point : float
        The tipping point between the two models.

    Returns
    -------
    callable
    """

    db_per_step = 20*numpy.log10(tipping_point)/tipping_point
    log_scaler = get_log_magnitude_scaling(scale_factor, db_per_step)

    def scaler(data):
        out = data/scale_factor
        above_tipping = (out > tipping_point)
        out[above_tipping] = log_scaler(data[above_tipping])
        return out
    return scaler


class ApplyAmplitudeScalingFunction(ComplexFormatFunction):
    __slots__ = ('_scaling_function', )
    _allowed_ordering = ('MP', 'PM')
    has_inverse = False

    def __init__(
            self,
            raw_dtype: Union[str, numpy.dtype],
            order: str,
            scaling_function: Optional[Callable] = None,
            raw_shape: Optional[Tuple[int, ...]] = None,
            formatted_shape: Optional[Tuple[int, ...]] = None,
            reverse_axes: Optional[Tuple[int, ...]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None,
            band_dimension: int = -1):
        """

        Parameters
        ----------
        raw_dtype : str|numpy.dtype
            The raw datatype. Valid options dependent on the value of order.
        order : str
            One of `('MP', 'PM')`, with allowable raw_dtype
            `('uint8', 'uint16', 'uint32', 'float32', 'float64')`.
        scaling_function : Optional[Callable]
        raw_shape : None|Tuple[int, ...]
        formatted_shape : None|Tuple[int, ...]
        reverse_axes : None|Tuple[int, ...]
        transpose_axes : None|Tuple[int, ...]
        band_dimension : int
            Which band is the complex dimension, **after** the transpose operation.
        """

        self._scaling_function = None
        ComplexFormatFunction.__init__(
            self, raw_dtype, order, raw_shape=raw_shape, formatted_shape=formatted_shape,
            reverse_axes=reverse_axes, transpose_axes=transpose_axes, band_dimension=band_dimension)
        self._set_scaling_function(scaling_function)

    @property
    def scaling_function(self) -> Optional[Callable]:
        """
        The magnitude scaling function.

        Returns
        -------
        None|Callable
        """

        return self._scaling_function

    def _set_scaling_function(self, value: Optional[Callable]):
        if value is None:
            self._scaling_function = None
            return
        if not isinstance(value, Callable):
            raise TypeError('scaling_function must be callable')
        self._scaling_function = value

    def _forward_magnitude_theta(
            self,
            data: numpy.ndarray,
            out: numpy.ndarray,
            magnitude: numpy.ndarray,
            theta: numpy.ndarray,
            subscript: Tuple[slice, ...]) -> None:
        if self._scaling_function is not None:
            magnitude = self._scaling_function(magnitude)
        ComplexFormatFunction._forward_magnitude_theta(
            self, data, out, magnitude, theta, subscript)


def _extract_transform_data(
        image_header: Union[ImageSegmentHeader, ImageSegmentHeader0],
        band_dimension: int):
    """
    Helper function for defining necessary transform_data definition for
    interpreting image segment data.

    Parameters
    ----------
    image_header : ImageSegmentHeader|ImageSegmentHeader0

    Returns
    -------
    None|str|callable
    """

    if len(image_header.Bands) != 2:
        raise ValueError('Got unhandled case of {} image bands'.format(len(image_header.Bands)))

    complex_order = image_header.Bands[0].ISUBCAT+image_header.Bands[1].ISUBCAT
    if complex_order not in ['IQ', 'QI', 'MP', 'PM']:
        raise ValueError('Got unhandled complex order `{}`'.format(complex_order))

    bpp = int(image_header.NBPP/8)
    pv_type = image_header.PVTYPE
    if pv_type == 'INT':
        raw_dtype = '>u{}'.format(bpp)
    elif pv_type == 'SI':
        raw_dtype = '>i{}'.format(bpp)
    elif pv_type == 'R':
        raw_dtype = '>f{}'.format(bpp)
    else:
        raise ValueError('Got unhandled PVTYPE {}'.format(pv_type))

    # noinspection PyUnresolvedReferences
    tre = None if image_header.ExtendedHeader.data is None else \
        image_header.ExtendedHeader.data['CMETAA']  # type: Optional[CMETAA]

    if tre is None:
        return ComplexFormatFunction(raw_dtype, complex_order, band_dimension=band_dimension)

    cmetaa = tre.DATA
    if cmetaa.CMPLX_PHASE_SCALING_TYPE.strip() != 'NS':
        raise ValueError(
            'Got unsupported CMPLX_PHASE_SCALING_TYPE {}'.format(
                cmetaa.CMPLX_PHASE_SCALING_TYPE))

    remap_type = cmetaa.CMPLX_MAG_REMAP_TYPE.strip()
    if remap_type == 'NS':
        if complex_order in ['IQ', 'QI']:
            return ComplexFormatFunction(raw_dtype, complex_order, band_dimension=band_dimension)
        else:
            raise ValueError(
                'Got unexpected state where cmetaa.CMPLX_MAG_REMAP_TYPE is "NS",\n\t '
                'but Band[0].ISUBCAT/Band[1].ISUBCAT = `{}`'.format(complex_order))
    elif remap_type not in ['LINM', 'LINP', 'LOGM', 'LOGP', 'LLM']:
        raise ValueError('Got unsupported CMETAA.CMPLX_MAG_REMAP_TYPE {}'.format(remap_type))

    if complex_order not in ['MP', 'PM']:
        raise ValueError(
            'Got unexpected state where cmetaa.CMPLX_MAG_REMAP_TYPE is `{}`,\n\t'
            'but Band[0].ISUBCAT/Band[1].ISUBCAT = `{}`'.format(
                remap_type, complex_order))

    scale_factor = float(cmetaa.CMPLX_LIN_SCALE)
    if remap_type == 'LINM':
        scaling_function = get_linear_magnitude_scaling(scale_factor)
    elif remap_type == 'LINP':
        scaling_function = get_linear_power_scaling(scale_factor)
    elif remap_type == 'LOGM':
        # NB: there is nowhere in the CMETAA structure to define
        #   the db_per_step value. Strangely, the use of this value is laid
        #   out in the STDI-0002 standards document, which defines CMETAA
        #   structure. We will generically use a value which maps the
        #   max uint8 value to the max int16 value.
        db_per_step = 300*numpy.log(2)/255.0
        scaling_function = get_log_magnitude_scaling(scale_factor, db_per_step)
    elif remap_type == 'LOGP':
        db_per_step = 300*numpy.log(2)/255.0
        scaling_function = get_log_power_scaling(scale_factor, db_per_step)
    elif remap_type == 'LLM':
        scaling_function = get_linlog_magnitude_scaling(
            scale_factor, int(cmetaa.CMPLX_LINLOG_TP))
    else:
        raise ValueError('Got unhandled CMETAA.CMPLX_MAG_REMAP_TYPE {}'.format(remap_type))
    return ApplyAmplitudeScalingFunction(raw_dtype, complex_order, scaling_function, band_dimension=band_dimension)


######
# The interpreter and reader objects

class ComplexNITFDetails(NITFDetails):
    """
    Details object for NITF file containing complex data.
    """

    __slots__ = (
        '_segment_status', '_segment_bands', '_sicd_meta', '_reverse_axes', '_transpose_axes')

    def __init__(
            self,
            file_name: str,
            reverse_axes: Union[None, int, Sequence[int]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None):
        """

        Parameters
        ----------
        file_name : str
            file name for a NITF file containing a complex SICD
        reverse_axes : None|Sequence[int]
            Any entries should be restricted to `{0, 1}`. The presence of
            `0` means to reverse the rows (in the raw sense), and the presence
            of `1` means to reverse the columns (in the raw sense).
        transpose_axes : None|Tuple[int, ...]
            If presented this should be only `(1, 0)`.
        """

        self._reverse_axes = reverse_axes
        self._transpose_axes = transpose_axes
        self._segment_status = None
        self._sicd_meta = None
        self._segment_bands = None
        NITFDetails.__init__(self, file_name)
        self._find_complex_image_segments()
        if len(self.sicd_meta) == 0:
            raise SarpyIOError(
                'No complex valued image segments found in file {}'.format(file_name))

    @property
    def reverse_axes(self) -> Union[None, int, Sequence[int]]:
        return self._reverse_axes

    @property
    def transpose_axes(self) -> Optional[Tuple[int, ...]]:
        return self._transpose_axes

    @property
    def segment_status(self) -> Tuple[bool, ...]:
        """
        Tuple[bool, ...]: Where each image segment is viable for use.
        """

        return self._segment_status

    @property
    def sicd_meta(self) -> Tuple[SICDType, ...]:
        """
        Tuple[SICDType, ...]: The best inferred sicd structures.
        """

        return self._sicd_meta

    @property
    def segment_bands(self) -> Tuple[Tuple[int, Optional[int]], ...]:
        """
        This describes the structure for the output data segments from the NITF,
        with each entry of the form `(image_segment, output_band)`, where
        `output_band` will be `None` if the image segment has exactly one
        complex band.

        Returns
        -------
        Tuple[Tuple[int, Optional[int]], ...]
            The band details for use.
        """

        return self._segment_bands

    def _check_band_details(
            self,
            index: int,
            sicd_meta: List,
            segment_status: List,
            segment_bands: List):
        if len(segment_status) != index:
            raise ValueError('Inconsistent status checking state')
        image_header = self.img_headers[index]
        if image_header.ICAT.strip() not in ['SAR', 'SARIQ']:
            segment_status.append(False)
            return

        # construct a preliminary sicd
        sicd = extract_sicd(image_header, self._transpose_axes is not None)
        bands = image_header.Bands
        pvtype = image_header.PVTYPE

        # handle odd bands
        if (len(bands) % 2) == 1:
            if image_header.PVTYPE != 'C':
                # it's not complex, so we're done
                segment_status.append(False)
                return
            segment_status.append(True)
            sicd_meta.append(sicd)
            segment_bands.append((index, len(bands)))
            return

        # we have an even number of bands - ensure that the bands are marked
        #  IQ/QI/MP/PM
        order = bands[0].ISUBCAT + bands[1].ISUBCAT
        if order not in ['IQ', 'QI', 'MP', 'PM']:
            segment_status.append(False)
            return

        if len(bands) == 2:
            # this should be the most common by far

            segment_status.append(True)
            sicd_meta.append(sicd)
            segment_bands.append((index, 1))
            return

        for i in range(2, len(bands), 2):
            if order != bands[i].ISUBCAT + bands[i+1].ISUBCAT:
                logging.error(
                    'Image segment appears to multiband with switch complex ordering')
                segment_status.append(False)
                return

        if order in ['IQ', 'QI']:
            if pvtype not in ['SI', 'R']:
                logging.error(
                    'Image segment appears to be complex of order `{}`, \n\t'
                    'but PVTYPE is `{}`'.format(order, pvtype))
                segment_status.append(False)

        if order in ['MP', 'PM']:
            if pvtype not in ['INT', 'R']:
                logging.error(
                    'Image segment appears to be complex of order `{}`, \n\t'
                    'but PVTYPE is `{}`'.format(order, pvtype))
                segment_status.append(False)

        segment_status.append(True)
        sicd_meta.append(sicd)
        segment_bands.append((index, int(len(bands)/2)))

    def _find_complex_image_segments(self):
        """
        Find complex image segments.

        Returns
        -------
        None
        """

        sicd_meta = []
        segment_status = []
        segment_bands = []
        for index in range(len(self.img_headers)):
            self._check_band_details(index, sicd_meta, segment_status, segment_bands)
        self._segment_status = tuple(segment_status)
        use_sicd_meta = []
        use_segment_bands = []
        for (the_index, out_bands), sicd in zip(segment_bands, sicd_meta):
            if out_bands == 1:
                use_sicd_meta.append(sicd)
                use_segment_bands.append((the_index, None))
            else:
                for j in range(out_bands):
                    use_sicd_meta.append(sicd.copy())
                    use_segment_bands.append((the_index, j))
        self._sicd_meta = tuple(use_sicd_meta)
        self._segment_bands = tuple(use_segment_bands)


class ComplexNITFReader(NITFReader, SICDTypeReader):
    """
    A reader for complex valued NITF elements, this should be explicitly tried AFTER
    the SICDReader.
    """

    def __init__(
            self,
            nitf_details: Union[str, ComplexNITFDetails],
            reverse_axes: Union[None, int, Sequence[int]] = None,
            transpose_axes: Optional[Tuple[int, ...]] = None):
        """

        Parameters
        ----------
        nitf_details : str|ComplexNITFDetails
        reverse_axes : None|Sequence[int]
            Any entries should be restricted to `{0, 1}`. The presence of
            `0` means to reverse the rows (in the raw sense), and the presence
            of `1` means to reverse the columns (in the raw sense).
        transpose_axes : None|Tuple[int, ...]
            If presented this should be only `(1, 0)`.
        """

        if isinstance(nitf_details, str):
            nitf_details = ComplexNITFDetails(
                nitf_details, reverse_axes=reverse_axes, transpose_axes=transpose_axes)
        if not isinstance(nitf_details, ComplexNITFDetails):
            raise TypeError('The input argument for ComplexNITFReader must be a filename or '
                            'ComplexNITFDetails object.')

        SICDTypeReader.__init__(self, None, nitf_details.sicd_meta)
        NITFReader.__init__(
            self,
            nitf_details,
            reader_type="SICD",
            reverse_axes=nitf_details.reverse_axes,
            transpose_axes=nitf_details.transpose_axes)
        self._check_sizes()

    @property
    def nitf_details(self) -> ComplexNITFDetails:
        """
        ComplexNITFDetails: The NITF details object.
        """

        # noinspection PyTypeChecker
        return self._nitf_details

    def get_nitf_dict(self):
        """
        Populate a dictionary with the pertinent NITF header information. This
        is for use in more faithful preservation of NITF header information
        in copying or rewriting sicd files.

        Returns
        -------
        dict
        """

        out = {}
        security = {}
        security_obj = self.nitf_details.nitf_header.Security
        # noinspection PyProtectedMember
        for field in NITFSecurityTags._ordering:
            value = getattr(security_obj, field).strip()
            if value != '':
                security[field] = value
        if len(security) > 0:
            out['Security'] = security

        out['OSTAID'] = self.nitf_details.nitf_header.OSTAID
        out['FTITLE'] = self.nitf_details.nitf_header.FTITLE
        return out

    def populate_nitf_information_into_sicd(self):
        """
        Populate some pertinent NITF header information into the SICD structure.
        This provides more faithful copying or rewriting options.
        """

        nitf_dict = self.get_nitf_dict()
        for sicd_meta in self._sicd_meta:
            sicd_meta.NITF = copy.deepcopy(nitf_dict)

    def depopulate_nitf_information(self):
        """
        Eliminates the NITF information dict from the SICD structure.
        """

        for sicd_meta in self._sicd_meta:
            sicd_meta.NITF = {}

    def get_format_function(
            self,
            raw_dtype: numpy.dtype,
            complex_order: Optional[str],
            lut: Optional[numpy.ndarray],
            band_dimension: int,
            image_segment_index: Optional[int] = None,
            **kwargs) -> Optional[FormatFunction]:
        image_header = self.nitf_details.img_headers[image_segment_index]
        bands = len(image_header.Bands)
        if complex_order is not None and bands == 2:
            return _extract_transform_data(image_header, band_dimension)
        # TODO: strange nonstandard float16 handling?
        return NITFReader.get_format_function(
            self, raw_dtype, complex_order, lut, band_dimension, image_segment_index, **kwargs)

    def _check_image_segment_for_compliance(
            self,
            index: int,
            img_header: Union[ImageSegmentHeader, ImageSegmentHeader0]) -> bool:
        return self.nitf_details.segment_status[index]

    def find_image_segment_collections(self) -> Tuple[Tuple[int, ...]]:
        return tuple((entry[0], ) for entry in self.nitf_details.segment_bands)

    def create_data_segment_for_collection_element(self, collection_index: int) -> DataSegment:
        the_index, the_band = self.nitf_details.segment_bands[collection_index]
        if the_index not in self._image_segment_data_segments:
            data_segment = self.create_data_segment_for_image_segment(the_index, apply_format=True)
        else:
            data_segment = self._image_segment_data_segments[the_index]

        if the_band is None:
            return data_segment
        else:
            return SubsetSegment(
                data_segment, (slice(None, None, 1), slice(None, None, 1), slice(the_band, the_band+1, 1)),
                'formatted', close_parent=True)


def final_attempt(file_name: str) -> Optional[ComplexNITFReader]:
    """
    Contingency check to open for some other complex NITF type file.
    Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str|BinaryIO
        the file_name to check

    Returns
    -------
    ComplexNITFReader|None
    """

    if is_file_like(file_name):
        return None

    try:
        nitf_details = ComplexNITFDetails(file_name)
        logger.info('File {} is determined to be some other format complex NITF.')
        return ComplexNITFReader(nitf_details)
    except (SarpyIOError, ValueError):
        return None
